import enum
from typing import List, Dict


EPSILON = 1.e-10


class ContractType(enum.Enum):
    """
    Enum class for contract type.
    """
    Stock = 1
    Option = 2
    Future = 3
    FX = 4
    Index = 5


class PriceBidAskType(enum.Enum):
    """
    Enum class for price bid-ask type.
    """
    Bid = 1
    Ask = 2
    Mid = 3


class PriceOHLCType(enum.Enum):
    """
    Enum class for price OHLC type.
    """
    Open = 1
    High = 2
    Low = 3
    Close = 4


class SignalInfo:
    """
    class for signal info.
    """
    def __init__(self,
                 signal_signature: str,
                 input_names: List[str],
                 contract_types: List[ContractType],
                 price_ohlc_types: List[PriceOHLCType],
                 warmup_length: int,
                 custom_params: Dict) -> None:
        """
        Initialize signal info

        Parameters
        ----------
        signal_signature: str
            Unique signature of the signal.
        input_names: List[str]
            List of inputs the signal is listening to.
        contract_types: List[ContractType]
            List of contract types.
        price_ohlc_types: List[PriceOHLCType]
            List of price OHLC types.
        warmup_length: int
            Warm-up length.
        custom_params: Dict
            Other signal specific parameters.
        """
        self._input_names = input_names
        self._contract_types = contract_types
        self._price_ohlc_types = price_ohlc_types
        self._warmup_length = warmup_length
        self._custom_params = custom_params
        self._signal_signature = signal_signature

    def get_input_names(self) -> List[str]:
        return self._input_names

    def get_contract_types(self) -> List[ContractType]:
        return self._contract_types

    def get_price_ohlc_types(self) -> List[PriceOHLCType]:
        return self._price_ohlc_types

    def get_warmup_length(self) -> int:
        return self._warmup_length

    def get_custom_params(self) -> Dict:
        return self._custom_params

    def get_signal_signature(self) -> str:
        return self._signal_signature


class StrategyInfo:
    """
    class for strategy info.
    """
    def __init__(self,
                 strategy_signature: str,
                 input_names: List[str],
                 warmup_length: int,
                 custom_params: Dict,
                 contract_names: List[str],
                 contract_types: List[ContractType] = None,
                 combo_definition: Dict[str, List[float]] = None) -> None:
        """
        Initialize strategy info

        Parameters
        ----------
        strategy_signature: str
            Unique signature of the signal.
        input_names: List[str]
            List of inputs the strategy is listening to.
        warmup_length: int
            Number of data points to 'burn'
        custom_params: Dict
            Other strategy specific parameters.
        contract_names: List[str]
            Contract names for TradeCombos creation.
        contract_types: List[ContractType]
            Contract types.
        combo_definition: Dict[str, List[float]]
            Weight dictionay for TradeCombos creation.
        """
        # Check input integrity
        if contract_types is not None:
            assert len(contract_names) == len(contract_types),\
                'Contract names and contract types have different lengths.'
        for k, v in combo_definition.items():
            assert len(v) == len(contract_names), 'Contract names and definition for ' + k + ' have different lengths.'

        self._input_names = input_names
        self._warmup_length = warmup_length
        self._custom_params = custom_params
        self._strategy_signature = strategy_signature
        self._contract_names = contract_names
        self._contract_types = contract_types
        self._combo_definition = combo_definition

    def get_input_names(self) -> List[str]:
        return self._input_names

    def get_warmup_length(self) -> int:
        return self._warmup_length

    def get_custom_params(self) -> Dict:
        return self._custom_params

    def get_contract_names(self) -> List[str]:
        return self._contract_names

    def get_contract_types(self) -> List[ContractType]:
        return self._contract_types

    def get_combo_definition(self) -> Dict[str, List[float]]:
        return self._combo_definition

    def get_strategy_signature(self) -> str:
        return self._strategy_signature



import asyncio
from contextlib import contextmanager
import datetime
from eventkit import Event
from typing import Dict, List
import yaml

from . import brokerinterface
from . import signal
from . import strategy
from . import utils


class LiveTrader:
    """
    Class for live trader.
    TBD: Make it a singleton
    """

    def __init__(self,
                 broker_api_signature: str,
                 signal_info: Dict[str, utils.SignalInfo],
                 strategy_info: Dict[str, utils.StrategyInfo],
                 path_to_control_file: str
                 ) -> None:
        """
        Initialize live trader.

        Parameters
        ----------
        broker_api_signature: str
            Broker API signature.
        signal_info: Dict[str, utils.SignalInfo]
            Information for building signals.
        strategy_info: Dict[str, utils.StrategyInfo]
            Information for building strategies.
        path_to_control_file: str
            File path to the control file.
        """
        self._broker = brokerinterface.BrokerAPIBase(broker_api_signature)
        self._signal_info = signal_info
        self._strategy_info = strategy_info
        self._strategy_dict = {}
        self.contract_dict = {}
        self._is_trading: bool = True
        self._path_to_control_file = path_to_control_file
        self._run_until: datetime.datetime = None

        # Create contracts
        self.contract_dict = self.create_contract_dict()

        self._portfolio_manager = brokerinterface.PortfolioManagerBase(self.get_broker().get_class_signature(),
                                                                       self.get_broker())
        self._order_manager = brokerinterface.OrderManagerBase(self.get_broker().get_class_signature(),
                                                               self.get_broker(), self._portfolio_manager,
                                                               self.contract_dict)

    def get_portfolio_manager(self) -> brokerinterface.PortfolioManagerBase:
        return self._portfolio_manager

    def get_broker(self) -> brokerinterface.BrokerAPIBase:
        return self._broker

    def disconnect(self) -> None:
        self.get_broker().disconnect()

    def get_is_trading(self) -> bool:
        return self._is_trading

    def stop_trading(self) -> None:
        self._is_trading = False

    @contextmanager
    def setup(self) -> None:
        self.get_broker().connect()
        # Create broker relays
        relay_dict = self.create_relay_dict()

        # Create signals
        data_event_dict: Dict[str, Event] = {}
        for key, value in relay_dict.items():
            data_event_dict[key] = value.get_event()
        signal_dict = self.create_signal_dict(data_event_dict)

        # Subscribe to data
        for name, contract in self.contract_dict.items():
            # TBD: Allow for bid or ask
            live_bar = self.get_broker().request_live_bars(contract, utils.PriceBidAskType.Mid)
            live_bar.updateEvent += relay_dict[name].live_bar

        # Create strategies
        strategy_dict = self.create_strategy_dict(signal_dict, data_event_dict, self._order_manager)
        self._strategy_dict = strategy_dict

        yield signal_dict, strategy_dict

    def start_trading(self) -> None:
        """
        Start live trading.
        """
        with self.setup() as (signal_dict, strategy_dict):
            event_loop = asyncio.get_event_loop()

            task = event_loop.create_task(self.async_sleep())
            event_loop.run_until_complete(task)

    async def async_sleep(self):
        while self.get_is_trading():
            # Check for new instruction
            with open(self._path_to_control_file, "r") as stream:
                try:
                    instruction_info = yaml.safe_load(stream)
                    self.carry_out_instruction(instruction_info)
                except yaml.YAMLError:
                    pass
            await asyncio.sleep(5)

    def carry_out_instruction(self, instruction_info) -> None:
        """Carry out the instruction as specified by the control file."""
        # Activity
        if instruction_info['activity'] is not None:
            if instruction_info['activity']['is_active'] is not None:
                if not instruction_info['activity']['is_active']:
                    # is_active has top priority
                    self._is_trading = False
            elif self._run_until is not None and self._run_until < datetime.datetime.now():
                # Time to shut down
                self._is_trading = False

            if ((instruction_info['activity']['run_until'] is not None)
                    and (instruction_info['activity']['seconds_to_shut_down'] is not None)):
                raise ValueError('LiveTrader.carry_out_instruction: run_until and seconds_to_shut_down '
                                 'cannot be specified simultaneously.')
            elif instruction_info['activity']['run_until'] is not None:
                self._run_until = datetime.strptime(instruction_info['activity']['run_until'], '%m/%d/%y %H:%M:%S')
            elif ((instruction_info['activity']['seconds_to_shut_down'] is not None)
                    and (self._run_until is None)):
                # We only consider seconds_to_shut_down at start up, otherwise we run into infinite loop
                datetime_now = datetime.datetime.now()
                self._run_until = datetime_now\
                    + datetime.timedelta(0, int(instruction_info['activity']['seconds_to_shut_down']))

        # Monitoring
        if instruction_info['monitoring'] is not None:
            if instruction_info['monitoring']['by_strategy'] is not None:
                if instruction_info['monitoring']['by_strategy'] == 'print':
                    print(self.get_portfolio_manager().get_portfolio_info())
                elif instruction_info['monitoring']['by_strategy'] == 'csv':
                    by_strategy_output_file = 'alpyen_portfolio_by_strategy.csv'
                    with open(by_strategy_output_file, 'a+') as f_bs:
                        f_bs.write('\n')
                        self.get_portfolio_manager().get_portfolio_info().to_csv(f_bs,
                                                                                 index=False,
                                                                                 mode='a',
                                                                                 sep='|')
            if instruction_info['monitoring']['account'] is not None:
                if instruction_info['monitoring']['account']:
                    account_output_file = 'alpyen_broker_account.csv'
                    with open(account_output_file, 'a+') as f_a:
                        f_a.write('\n')
                        self.get_broker().get_account_info().to_csv(f_a,
                                                                    index=False,
                                                                    mode='a',
                                                                    sep='|')
            if instruction_info['monitoring']['portfolio'] is not None:
                if instruction_info['monitoring']['portfolio']:
                    portfolio_output_file = 'alpyen_portfolio.csv'
                    with open(portfolio_output_file, 'a+') as f_p:
                        f_p.write('\n')
                        self.get_broker().get_portfolio_info().to_csv(f_p,
                                                                      index=False,
                                                                      mode='a',
                                                                      sep='|')

        # Strategies
        if instruction_info['strategies'] is not None:
            for strategy_name, dict_for_strategy in instruction_info['strategies'].items():
                for strategy_attribute, value in dict_for_strategy.items():
                    if strategy_attribute == 'on_off':
                        self.switch_strategy_on_off(strategy_name, value)

    def create_contract_dict(self) -> Dict[str, brokerinterface.BrokerContractBase]:
        """Create contract dictionary."""
        output_dict: Dict[str, brokerinterface.BrokerContractBase] = {}
        for k_signal, v_signal in self._signal_info.items():
            for contract_type, input_name in zip(v_signal.get_contract_types(), v_signal.get_input_names()):
                if input_name not in output_dict:
                    output_dict[input_name] = brokerinterface.BrokerContractBase(
                        self.get_broker().get_class_signature(), contract_type, input_name)

        # In case there are traded contracts that are not in any signal
        strategy_info_dict: Dict[str, utils.StrategyInfo] = self._strategy_info
        for k, v in strategy_info_dict.items():
            for i in range(len(v.get_contract_names())):
                if v.get_contract_names()[i] not in output_dict:
                    output_dict[v.get_contract_names()[i]] = brokerinterface.BrokerContractBase(
                        self.get_broker().get_class_signature(), v.get_contract_types()[i], v.get_contract_names()[i])
        return output_dict

    def create_relay_dict(self) -> Dict[str, brokerinterface.BrokerEventRelayBase]:
        """Create relay dictionary."""
        output_dict: Dict[str, brokerinterface.BrokerEventRelayBase] = {}
        for k_signal, v_signal in self._signal_info.items():
            for price_type, input_name in zip(v_signal.get_price_ohlc_types(), v_signal.get_input_names()):
                if input_name not in output_dict:
                    output_dict[input_name] = brokerinterface.BrokerEventRelayBase(
                        self.get_broker().get_class_signature(), input_name, price_type)

        # In case there are traded contracts that are not in any signal
        strategy_info_dict: Dict[str, utils.StrategyInfo] = self._strategy_info
        for k, v in strategy_info_dict.items():
            for i in range(len(v.get_contract_names())):
                if v.get_contract_names()[i] not in output_dict:
                    # OK to default to close price for mtm
                    output_dict[v.get_contract_names()[i]] = brokerinterface.BrokerEventRelayBase(
                        self.get_broker().get_class_signature(), v.get_contract_names()[i], utils.PriceOHLCType.Close)
        return output_dict

    def create_signal_dict(self,
                           data_event_dict: Dict[str, Event]) -> Dict[str, signal.SignalBase]:
        """
        Create signal dictionary.

        Parameters
        ----------
        data_event_dict: Dict[str, Event]
            Data event dictionary.
        """
        output_dict: Dict[str, signal.SignalBase] = {}
        signal_info_dict: Dict[str, utils.SignalInfo] = self._signal_info
        for k, v in signal_info_dict.items():
            price_event_list: List[Event] = []
            for name_i in v.get_input_names():
                price_event_list.append(data_event_dict.get(name_i))
            # Create signal
            my_signal = signal.SignalBase(v.get_signal_signature(),
                                          price_event_list,
                                          v.get_warmup_length(),
                                          **v.get_custom_params())
            output_dict[my_signal.get_signal_name()] = my_signal
        return output_dict

    def create_strategy_dict(self,
                             signal_dict: Dict[str, signal.SignalBase],
                             data_event_dict: Dict[str, Event],
                             order_manager: brokerinterface.OrderManagerBase) -> Dict[str, strategy.StrategyBase]:
        """
        Create strategy dictionary.

        Parameters
        ----------
        signal_dict: Dict[str, signal.SignalBase]
            Signal dictionary.
        data_event_dict: Dict[str, Event]
            Data event dictionary.
        order_manager: brokerinterface.OrderManagerBase
            Order manager object.
        """
        output_dict: Dict[str, strategy.StrategyBase] = {}
        strategy_info_dict: Dict[str, utils.StrategyInfo] = self._strategy_info
        for k, v in strategy_info_dict.items():
            price_event_list: List[Event] = []
            for i in range(len(v.get_contract_names())):
                price_event_list.append(data_event_dict.get(v.get_contract_names()[i]))
            # Create TradeCombos
            trade_combos = strategy.TradeCombos(price_event_list, v.get_combo_definition())
            # Create strategy
            signal_event_list: List[Event] = []
            for i in range(len(v.get_input_names())):
                signal_event_list.append(signal_dict.get(v.get_input_names()[i]).get_signal_event())
            my_strategy = strategy.StrategyBase(v.get_strategy_signature(), signal_event_list,
                                                trade_combos, v.get_warmup_length(), **v.get_custom_params(),
                                                order_manager=order_manager)
            output_dict[my_strategy.get_strategy_name()] = my_strategy
        return output_dict

    def switch_all_strategies_on_off(self, activity: bool) -> None:
        """
        Turn on / off all strategies.

        Parameters
        ----------
        activity: bool
            True to turn on; False to turn off.
        """
        for value in self._strategy_dict.values():
            value.set_strategy_active(activity)

    def switch_strategy_on_off(self, strategy_name: str, activity: bool) -> None:
        """
        Turn on / off a specific strategy.

        Parameters
        ----------
        strategy_name: str
            Name of the strategy.
        activity: bool
            True to turn on; False to turn off.
        """
        self._strategy_dict[strategy_name].set_strategy_active(activity)

    def liquidate_all(self) -> None:
        """
        Liquidate all holdings (only those holdings portfolio manager is aware of).

        Note that:
        - All strategies would be switched off
        - There is no guarantee that the book would be clean after running this function
            - Some positions can be outside alpyen
            - Some positions can be dangling order when this function is triggered
        """
        current_holdings = self._portfolio_manager.contract_info_df

        # Switch off all strategies
        self.switch_all_strategies_on_off(False)

        # Go through the holdings and do unwind trades
        for _, holding in current_holdings.iterrows():
            # Retrieve contract
            contract_to_trade = self.contract_dict[holding['symbol'].iloc[0]]
            trade_object = self._order_manager.order_wrapper(contract_to_trade, -holding['position'].iloc[0])

        # Clear all records in portfolio manager
        self._portfolio_manager.portfolio_info_df = self._portfolio_manager.portfolio_info_df[0:0]
        self._portfolio_manager.contract_info_df = self._portfolio_manager.contract_info_df[0:0]



# This should be the only file that accesses broker api
from abc import abstractmethod
from datetime import date
from eventkit import Event
import ib_insync as ibi  # For Interactive Brokers (IB)
import pandas as pd
from typing import Optional, Dict, List

from . import datacontainer
from . import utils


class BrokerEventRelayBase:
    """
    Base class for broker event relay.
    """
    _broker_relay_classes_registry = {}

    _broker_relay_signature = None

    @classmethod
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if cls._broker_relay_signature is None:
            raise KeyError('BrokerEventRelayBase: Missing signature for ' + str(cls))
        elif cls._broker_relay_signature in cls._broker_relay_classes_registry:
            raise KeyError('BrokerEventRelayBase: Conflict in signature ' + cls._broker_relay_signature)
        else:
            cls._broker_relay_classes_registry[cls._broker_relay_signature] = cls

    @classmethod
    def get_class_registry(cls):
        return cls._broker_relay_classes_registry

    def __new__(cls,
                signature_str: str,
                data_name: str,
                ohlc: utils.PriceOHLCType = utils.PriceOHLCType.Close
                ):
        """
        Initialize broker event relay.

        Parameters
        ----------
        signature_str: str
            Unique signature of the relay class.
        data_name: str
            Name of the input data.
        ohlc: utils.PriceOHLCType
            OHLC name (open, high, low, close, volume, etc.).
        """
        if signature_str not in cls.get_class_registry():
            raise ValueError('BrokerEventRelayBase.__new__: ' + signature_str + ' is not a valid key.')

        my_relay_obj = super().__new__(cls.get_class_registry()[signature_str])

        my_relay_obj._relay_event = Event(data_name)
        my_relay_obj._ohlc = ohlc
        my_relay_obj._data_name = data_name

        return my_relay_obj

    def __init__(self,
                 signature_str: str,
                 data_name: str,
                 ohlc: utils.PriceOHLCType = utils.PriceOHLCType.Close) -> None:
        pass

    def get_event(self) -> Event:
        return self._relay_event


class BrokerContractBase:
    """Base class for contract."""
    _broker_contract_classes_registry = {}

    _broker_contract_signature = None

    @classmethod
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if cls._broker_contract_signature is None:
            raise KeyError('BrokerContractBase: Missing signature for ' + str(cls))
        elif cls._broker_contract_signature in cls._broker_contract_classes_registry:
            raise KeyError('BrokerContractBase: Conflict in signature ' + cls._broker_contract_signature)
        else:
            cls._broker_contract_classes_registry[cls._broker_contract_signature] = cls

    @classmethod
    def get_class_registry(cls):
        return cls._broker_contract_classes_registry

    def __new__(cls,
                signature_str: str,
                type_: utils.ContractType,
                symbol: str,
                strike: Optional[float] = None,
                expiry: Optional[date] = None):
        """
        Initialize broker contract.

        Parameters
        ----------
        signature_str: str
            Unique signature of the contract class.
        type_: ContractType
            Contract type.
        symbol: str
            Ticker symbol.
        strike: Optional[float]
            Strike (optional).
        expiry: Optional[date]
            Expiry (optional).
        """
        if signature_str not in cls.get_class_registry():
            raise ValueError('BrokerContractBase.__new__: ' + signature_str + ' is not a valid key.')

        my_contract_obj = super().__new__(cls.get_class_registry()[signature_str])

        my_contract_obj._type = my_contract_obj.type_translation(type_)
        my_contract_obj._symbol = symbol
        my_contract_obj._strike = strike
        my_contract_obj._expiry = expiry
        my_contract_obj._contract = my_contract_obj.create_contract()

        return my_contract_obj

    def __init__(self,
                 signature_str: str,
                 type_: utils.ContractType,
                 symbol: str,
                 strike: Optional[float] = None,
                 expiry: Optional[date] = None) -> None:
        pass

    @abstractmethod
    def create_contract(self):
        pass

    @abstractmethod
    def type_translation(self, type_: utils.ContractType) -> str:
        pass

    def get_contract(self):
        return self._contract


class BrokerAPIBase:
    """Base class for broker API handle."""
    _broker_api_classes_registry = {}

    _broker_api_signature = None

    @classmethod
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if cls._broker_api_signature is None:
            raise KeyError('BrokerAPIBase: Missing signature for ' + str(cls))
        elif cls._broker_api_signature in cls._broker_api_classes_registry:
            raise KeyError('BrokerAPIBase: Conflict in signature ' + cls._broker_api_signature)
        else:
            cls._broker_api_classes_registry[cls._broker_api_signature] = cls

    @classmethod
    def get_class_registry(cls):
        return cls._broker_api_classes_registry

    def get_class_signature(self):
        return self._broker_api_signature

    def __new__(cls,
                signature_str: str):
        if signature_str not in cls.get_class_registry():
            raise ValueError('BrokerAPIBase.__new__: ' + signature_str + ' is not a valid key.')

        my_api_obj = super().__new__(cls.get_class_registry()[signature_str])
        return my_api_obj

    def __init__(self,
                 signature_str: str) -> None:
        pass

    def get_handle(self):
        return self._handle

    @abstractmethod
    def request_live_bars(self,
                          contract: BrokerContractBase,
                          price_type: utils.PriceBidAskType):
        pass

    @abstractmethod
    def connect(self, **kwargs) -> None:
        pass

    @abstractmethod
    def disconnect(self) -> None:
        pass

    @abstractmethod
    def get_account_info(self) -> pd.DataFrame:
        pass

    @abstractmethod
    def get_portfolio_info(self) -> pd.DataFrame:
        pass


class IBBrokerAPI(BrokerAPIBase):
    """Class for IB API handle."""

    _broker_api_signature = 'IB'

    def __init__(self, signature_str: str) -> None:
        ibi.util.startLoop()
        self._handle = ibi.IB()

    def connect(self,
                address: str = '127.0.0.1',
                port: int = 4002,
                client_id: int = 1) -> None:
        self.get_handle().connect(address, port, clientId=client_id)

    async def async_connect(self,
                address: str = '127.0.0.1',
                port: int = 4002,
                client_id: int = 1) -> None:
        await self.get_handle().connectAsync(address, port, clientId=client_id)

    def disconnect(self):
        self.get_handle().disconnect()

    def get_account_info(self) -> pd.DataFrame:
        output_df = pd.DataFrame(columns=['Net Value', 'Margin Requirement', 'Buying Power'])
        account_value_list = self.get_handle().accountSummary()
        temp_dict = {}
        for account_value_item in account_value_list:
            if account_value_item.tag == 'NetLiquidation':
                temp_dict['Net Value'] = account_value_item.value
            elif account_value_item.tag == 'MaintMarginReq':
                temp_dict['Margin Requirement'] = account_value_item.value
            elif account_value_item.tag == 'BuyingPower':
                temp_dict['Buying Power'] = account_value_item.value
        output_df = output_df.append(temp_dict, ignore_index=True)
        return output_df

    def get_portfolio_info(self) -> pd.DataFrame:
        output_df = pd.DataFrame(columns=['Security', 'Amount', 'Avg Cost',
                                          'Mkt Price', 'Realized PnL', 'Unrealized PnL'])
        my_portfolio = self.get_handle().portfolio()
        for portfolio_item in my_portfolio:
            temp_dict = {}
            temp_dict['Security'] = portfolio_item.contract.symbol
            temp_dict['Amount'] = portfolio_item.position
            temp_dict['Avg Cost'] = portfolio_item.averageCost
            temp_dict['Mkt Price'] = portfolio_item.marketPrice
            temp_dict['Realized PnL'] = portfolio_item.realizedPNL
            temp_dict['Unrealized PnL'] = portfolio_item.unrealizedPNL
            output_df = output_df.append(temp_dict, ignore_index=True)
        return output_df

    class IBBrokerEventRelay(BrokerEventRelayBase):
        """IB event relay"""
        _broker_relay_signature = 'IB'

        def __init__(self,
                     signature_str: str,
                     data_name: str,
                     ohlc: utils.PriceOHLCType = utils.PriceOHLCType.Close
                     ) -> None:
            """
            Initialize IB event relay.

            Parameters
            ----------
            signature_str: str
                Unique signature of the relay class.
            data_name: str
                Name of the input data.
            ohlc: utils.PriceOHLCType
                OHLC name (open, high, low, close, volume, etc.).
            """
            pass

        # TBD: Add different relay member functions (open, high, low, close, volume)
        def live_bar(self,
                     bars: ibi.RealTimeBarList,
                     has_new_bar: bool) -> None:
            """
            Translate IB real time bar event into price update.

            Parameters
            ----------
            bars: ibi.RealTimeBarList
                IB RealTimeBarList.
            has_new_bar: bool
                Whether there is new bar.
            """
            if has_new_bar:
                if self._ohlc == utils.PriceOHLCType.Close:
                    field = bars[-1].close
                else:
                    raise TypeError('IBBrokerEventRelay.live_bar: Unsupported data field type.')
                relay_data = datacontainer.TimeDouble(self._data_name, bars[-1].time, field)
                self._relay_event.emit(relay_data)

    class IBBrokerContract(BrokerContractBase):
        """Class for IB contracts."""
        _broker_contract_signature = 'IB'

        def __init__(self,
                     signature_str: str,
                     type_: utils.ContractType,
                     symbol: str,
                     strike: Optional[float] = None,
                     expiry: Optional[date] = None) -> None:
            pass

        def create_contract(self):
            if self._type == self.type_translation(utils.ContractType.FX):
                return ibi.contract.Forex(self._symbol)
            else:
                return ibi.contract.Contract(symbol=self._symbol,
                                             secType=self._type,
                                             lastTradeDateOrContractMonth='' if self._expiry is None
                                             else self._expiry.strftime('%Y%m%d'),
                                             strike=0.0 if self._strike is None else self._strike)

        def type_translation(self, type_: utils.ContractType) -> str:
            if type_ == utils.ContractType.Stock:
                return 'STK'
            elif type_ == utils.ContractType.Option:
                return 'OPT'
            elif type_ == utils.ContractType.FX:
                return 'CASH'
            elif type_ == utils.ContractType.Future:
                return 'FUT'
            elif type_ == utils.ContractType.Index:
                return 'IND'
            else:
                raise ValueError('IBBrokerContract.type_translation: Type not implemented.')

    def request_live_bars(self,
                          contract: IBBrokerContract,
                          price_type: utils.PriceBidAskType):
        """
        Request live price data bars.

        Parameters
        ----------
        contract: IBBrokerContract
            IB contract.
        price_type: utils.PriceBidAskType
            Price type.
        """
        ib_price_type_dict: Dict[utils.PriceBidAskType, str] = {utils.PriceBidAskType.Bid: 'BID',
                                                                utils.PriceBidAskType.Ask: 'ASK',
                                                                utils.PriceBidAskType.Mid: 'MIDPOINT'}
        return self.get_handle().reqRealTimeBars(contract.get_contract(), 5, ib_price_type_dict[price_type], False)


class PortfolioManagerBase:
    """
    Base class for portfolio manager.

    This class is responsible for:
    * Keeping an up-to-date record of security holdings
    * Keeping an up-to-date record of strategy-based holding
    * Reconciling security holdings with broker account record
    * Allow OrderManager to report new transaction
    * Provide holding info for strategies to decide on whether to trade
    """

    _broker_portfolio_manager_classes_registry = {}

    _broker_portfolio_manager_signature = None

    @classmethod
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if cls._broker_portfolio_manager_signature is None:
            raise KeyError('PortfolioManagerBase: Missing signature for ' + str(cls))
        elif cls._broker_portfolio_manager_signature in cls._broker_portfolio_manager_classes_registry:
            raise KeyError('PortfolioManagerBase: Conflict in signature ' + cls._broker_portfolio_manager_signature)
        else:
            cls._broker_portfolio_manager_classes_registry[cls._broker_portfolio_manager_signature] = cls

    @classmethod
    def get_class_registry(cls):
        return cls._broker_portfolio_manager_classes_registry

    def __new__(cls,
                signature_str: str,
                broker_api: BrokerAPIBase):
        """
        Initialize portfolio manager.

        Parameters
        ----------
        signature_str: str
            Unique signature of the portfolio manager class.
        broker_api: BrokerAPIBase
            Broker API.
        """
        if signature_str not in cls.get_class_registry():
            raise ValueError('PortfolioManagerBase.__new__: ' + signature_str + ' is not a valid key.')

        my_portfolio_manager_obj = super().__new__(cls.get_class_registry()[signature_str])

        my_portfolio_manager_obj._broker_handle = broker_api.get_handle()

        # Create dataframes for storing information
        my_portfolio_manager_obj.portfolio_info_df = pd.DataFrame(columns=['strategy_name', 'combo_name',
                                                                           'combo_position', 'combo_entry_price',
                                                                           'combo_mtm_price', 'unrealized_pnl',
                                                                           'realized_pnl'])
        my_portfolio_manager_obj.contract_info_df = pd.DataFrame(columns=['symbol', 'type', 'exchange',
                                                                          'currency', 'position'])

        return my_portfolio_manager_obj

    def __init__(self,
                 signature_str: str,
                 broker_api: BrokerAPIBase) -> None:
        pass

    def update_combo_mtm_price(self, combo_mtm: pd.DataFrame):
        """
        Update the combo MTM price.

        Parameters
        ----------
        combo_mtm: pd.DataFrame
            A dataframe with columns ['strategy_name', 'combo_name', 'combo_mtm_price'].
        """
        if len(self.portfolio_info_df) == 0:
            return
        temp_df = pd.merge(self.portfolio_info_df, combo_mtm,
                           how="left", on=['strategy_name', 'combo_name'])
        self.portfolio_info_df['combo_mtm_price'] = temp_df['combo_mtm_price_y'].fillna(temp_df['combo_mtm_price_x'])
        self.portfolio_info_df['unrealized_pnl'] = ((self.portfolio_info_df['combo_mtm_price']
                                                    - self.portfolio_info_df['combo_entry_price'])
                                                    * self.portfolio_info_df['combo_position'])

    def get_combo_position(self,
                           strategy_name: str,
                           combo_name: str) -> float:
        """
        Return the current position of a specific combo.

        Parameters
        ----------
        strategy_name: str
            Name of the strategy
        combo_name: str
            Name of the combo. Each combo in a strategy should have a unique name.

        Returns
        -------
            float
                The combo position.
        """
        is_existing = ((self.portfolio_info_df['strategy_name'] == strategy_name) &
                       (self.portfolio_info_df['combo_name'] == combo_name)).any()
        if is_existing:
            return (self.portfolio_info_df.loc[(self.portfolio_info_df['strategy_name'] == strategy_name) &
                                               (self.portfolio_info_df['combo_name'] == combo_name),
                                               'combo_position'].iloc[0])
        else:
            return 0.0

    def get_portfolio_info(self) -> pd.DataFrame:
        return self.portfolio_info_df

    @abstractmethod
    def portfolio_update(self, **kwargs) -> None:
        pass

    @abstractmethod
    def register_contract_trade(self, **kwargs) -> None:
        """
        Register completed contract trade from order manager
        """
        pass

    @abstractmethod
    def register_combo_trade(self, **kwargs) -> None:
        """
        Register completed combo trade from order manager
        """
        pass


class IBPortfolioManager(PortfolioManagerBase):
    """IB portfolio manager."""
    _broker_portfolio_manager_signature = 'IB'

    def __init__(self,
                 signature_str: str,
                 broker_api: IBBrokerAPI) -> None:
        """
        Initialize portfolio manager.

        Parameters
        ----------
        signature_str: str
            Unique signature of the portfolio manager class.
        broker_api: IBBrokerAPI
            IB Broker API.
        """
        pass

    def portfolio_update(self, ) -> None:
        pass

    def register_contract_trade(self,
                                symbol: str,
                                type_: str,
                                exchange: str,
                                currency: str,
                                position: float) -> None:
        """
        Register completed contract trade from order manager

        symbol: str
            Ticker symbol.
        type_: str
            Contract type (stock, future, etc.).
        exchange: str
            Exchange traded on.
        currency: str
            Denomination currency.
        position: float
            Amount traded
        """
        if ((self.contract_info_df['symbol'] == symbol) &
                (self.contract_info_df['type'] == type_) &
                (self.contract_info_df['exchange'] == exchange) &
                (self.contract_info_df['currency'] == currency)).any():
            self.contract_info_df.loc[(self.contract_info_df['symbol'] == symbol) &
                                      (self.contract_info_df['type'] == type_) &
                                      (self.contract_info_df['exchange'] == exchange) &
                                      (self.contract_info_df['currency'] == currency), 'position'] += position
        else:
            self.contract_info_df = self.contract_info_df.append({'symbol': symbol,
                                                                  'type': type_,
                                                                  'exchange': exchange,
                                                                  'currency': currency,
                                                                  'position': position},
                                                                 ignore_index=True)

    def register_combo_trade(self,
                             strategy_name: str,
                             combo_name: str,
                             combo_unit: float,
                             combo_def: List[float],
                             combo_entry_price: float) -> None:
        """
        Register completed combo trade from order manager.

        strategy_name: str
            Name of the strategy placing the order.
        combo_name: str
            Name of the combo. Each combo in a strategy should have a unique name.
        combo_unit: float
            Number of combo units traded.
        combo_def: List[float]
            Weight that defines the combo.
        combo_entry_price: float
            Entry price.
        """
        is_existing = ((self.portfolio_info_df['strategy_name'] == strategy_name) &
                       (self.portfolio_info_df['combo_name'] == combo_name)).any()
        if is_existing:
            existing_entry_price = (self.portfolio_info_df.loc[
                (self.portfolio_info_df['strategy_name'] == strategy_name) &
                (self.portfolio_info_df['combo_name'] == combo_name),
                'combo_entry_price'].iloc[0])
            existing_position = (self.portfolio_info_df.loc[
                (self.portfolio_info_df['strategy_name'] == strategy_name) &
                (self.portfolio_info_df['combo_name'] == combo_name),
                'combo_position'].iloc[0])
            # Update entry price
            if existing_position * combo_unit < 0.0:
                # If new order and existing holding are of opposite sign
                if abs(combo_unit) > abs(existing_position):
                    (self.portfolio_info_df.loc[(self.portfolio_info_df['strategy_name'] == strategy_name) &
                                                (self.portfolio_info_df['combo_name'] == combo_name),
                                                'combo_entry_price']) = combo_entry_price
                else:
                    # Do not update entry price, because old position not completely consumed.
                    pass
            else:
                # If new order and existing holding of same sign: Calculate weighted entry price
                new_weighted_entry_price = ((abs(existing_position) * existing_entry_price
                                             + abs(combo_unit) * combo_entry_price)
                                            / abs(existing_position + combo_unit))
                (self.portfolio_info_df.loc[(self.portfolio_info_df['strategy_name'] == strategy_name) &
                                            (self.portfolio_info_df['combo_name'] == combo_name),
                                            'combo_entry_price']) = new_weighted_entry_price
            # Update realized PnL
            position_closed = 0.0
            if existing_position * combo_unit < 0.0:
                # If new order and existing holding are of opposite sign
                position_closed = min(abs(existing_position), abs(combo_unit))
            existing_pnl = (self.portfolio_info_df.loc[(self.portfolio_info_df['strategy_name'] == strategy_name) &
                                                       (self.portfolio_info_df['combo_name'] == combo_name),
                                                       'realized_pnl']).iloc[0]

            existing_holding_direction = (existing_position / abs(existing_position)
                                          if abs(existing_position) > utils.EPSILON else 0.0)

            (self.portfolio_info_df.loc[(self.portfolio_info_df['strategy_name'] == strategy_name) &
                                        (self.portfolio_info_df['combo_name'] == combo_name),
                                        'realized_pnl']) = existing_pnl + existing_holding_direction * position_closed\
                * (combo_entry_price - existing_entry_price)

            # Update combo holding
            (self.portfolio_info_df.loc[(self.portfolio_info_df['strategy_name'] == strategy_name) &
                                        (self.portfolio_info_df['combo_name'] == combo_name),
                                        'combo_position']) = existing_position + combo_unit
        else:
            self.portfolio_info_df = self.portfolio_info_df.append({'strategy_name': strategy_name,
                                                                    'combo_name': combo_name,
                                                                    'combo_position': combo_unit,
                                                                    'combo_entry_price': combo_entry_price,
                                                                    'combo_mtm_price': combo_entry_price,
                                                                    'realized_pnl': 0.0},
                                                                   ignore_index=True)


class OrderManagerBase:
    """
    Base class for order manager.

    This class is responsible for:
    * Receiving orders from strategies
    * Sending orders to broker
    * Receiving confirmation from broker
    * Registering confirmed trades with PortfolioManager
    * Keeping track of dangling (i.e. unconfirmed) orders
    """

    _broker_order_manager_classes_registry = {}

    _broker_order_manager_signature = None

    @classmethod
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if cls._broker_order_manager_signature is None:
            raise KeyError('OrderManagerBase: Missing signature for ' + str(cls))
        elif cls._broker_order_manager_signature in cls._broker_order_manager_classes_registry:
            raise KeyError('OrderManagerBase: Conflict in signature ' + cls._broker_order_manager_signature)
        else:
            cls._broker_order_manager_classes_registry[cls._broker_order_manager_signature] = cls

    @classmethod
    def get_class_registry(cls):
        return cls._broker_order_manager_classes_registry

    def __new__(cls,
                signature_str: str,
                broker_api: BrokerAPIBase,
                portfolio_manager: PortfolioManagerBase,
                event_contract_dict: Dict[str, BrokerContractBase]):
        """
        Initialize order manager.

        Parameters
        ----------
        signature_str: str
            Unique signature of the order manager class.
        broker_api: BrokerAPIBase
            Broker API.
        portfolio_manager: PortfolioManagerBase
            Portfolio manager.
        event_contract_dict: Dict[str, BrokerContractBase]
            A dictionary allowing for mapping to contracts.
        """
        if signature_str not in cls.get_class_registry():
            raise ValueError('OrderManagerBase.__new__: ' + signature_str + ' is not a valid key.')

        my_order_manager_obj = super().__new__(cls.get_class_registry()[signature_str])

        my_order_manager_obj._broker_handle = broker_api.get_handle()
        my_order_manager_obj._portfolio_manager = portfolio_manager

        # Create dataframes for storing information
        my_order_manager_obj.order_info_df = pd.DataFrame(columns=['strategy_name', 'combo_name', 'contract_index',
                                                                   'combo_unit', 'dangling_order', 'entry_price',
                                                                   'order_id', 'time_stamp'])

        my_order_manager_obj._strategy_contracts: Dict[
            str, List[BrokerContractBase]] = {}  # A dictionary { strategy_name: contract_array }
        my_order_manager_obj._combo_weight: Dict[
            (str, str), List[float]] = {}  # A dictionary { (strategy_name, combo_name): combo_weight }

        my_order_manager_obj._event_contract_dict = event_contract_dict

        return my_order_manager_obj

    def __init__(self,
                 signature_str: str,
                 broker_api: IBBrokerAPI,
                 portfolio_manager: IBPortfolioManager,
                 event_contract_dict: Dict[str, IBBrokerAPI.IBBrokerContract]) -> None:
        pass

    def get_event_contract_dict(self):
        return self._event_contract_dict

    def get_portfolio_manager(self):
        return self._portfolio_manager

    @abstractmethod
    def place_order(self,
                    strategy_name: str,
                    combo_name: str,
                    time_stamp: str,
                    contract_index: int,
                    contract: BrokerContractBase,
                    unit: float
                    ) -> None:
        """
        Place order that is requested by strategy.

        Parameters
        ----------
        strategy_name: str
            Name of the strategy placing the order.
        combo_name: str
            Name of the combo. Each combo in a strategy should have a unique name.
        time_stamp: str
            Timestamp.
        contract_index: int
            Index of the contract as found in contract array.
        contract: BrokerContractBase
            Contract to be traded.
        unit: float
            Number of unit to trade.
        """
        pass

    @abstractmethod
    def register_combo_level_info(self,
                                  strategy_name: str,
                                  contract_array: List[BrokerContractBase],
                                  weight_array: List[float],
                                  combo_name: str) -> None:
        """
        Register combo level information that does not show up in place_order.

        Parameters
        ----------
        strategy_name: str
            Name of the strategy placing the order.
        contract_array: List[BrokerContractBase]
            List of contracts to be traded.
        weight_array: List[float]
            Weights of the contracts to be traded.
        combo_name: str
            Name of the combo. Each combo in a strategy should have a unique name.
        """
        pass

    @abstractmethod
    def update_order_status(self,
                            **kwargs) -> None:
        pass

    @abstractmethod
    def order_wrapper(self, **kwargs):
        pass


class IBOrderManager(OrderManagerBase):
    """IB order manager."""

    _broker_order_manager_signature = 'IB'

    def __init__(self,
                 signature_str: str,
                 broker_api: IBBrokerAPI,
                 portfolio_manager: IBPortfolioManager,
                 event_contract_dict: Dict[str, IBBrokerAPI.IBBrokerContract]) -> None:
        """
        Initialize order manager.

        Parameters
        ----------
        signature_str: str
            Unique signature of the order manager class.
        broker_api: IBBrokerAPI
            IB Broker API.
        portfolio_manager: IBPortfolioManager
            IB portfolio manager.
        event_contract_dict: Dict[str, IBBrokerAPI.IBBrokerContract]
            A dictionary allowing for mapping to contracts.
        """
        pass

    def register_combo_level_info(self,
                                  strategy_name: str,
                                  contract_array: List[IBBrokerAPI.IBBrokerContract],
                                  weight_array: List[float],
                                  combo_name: str) -> None:
        """
        Register combo level information that does not show up in place_order.

        Parameters
        ----------
        strategy_name: str
            Name of the strategy placing the order.
        contract_array: List[IBBrokerAPI.IBBrokerContract]
            List of contracts to be traded.
        weight_array: List[float]
            Weights of the contracts to be traded.
        combo_name: str
            Name of the combo. Each combo in a strategy should have a unique name.
        """
        self._strategy_contracts[strategy_name] = contract_array
        self._combo_weight[(strategy_name, combo_name)] = weight_array

    def order_wrapper(self,
                      contract: IBBrokerAPI.IBBrokerContract,
                      order_amount: float,
                      order_type: str = 'MKT'):
        """
        Wrapper for the broker's order function.

        Parameters
        ----------
        contract: IBBrokerAPI.IBBrokerContract
            Contract to be traded.
        order_amount: float
            Amount to trade.
        order_type: str
            Order type.
        """
        buy_sell: str = 'BUY' if order_amount > utils.EPSILON else 'SELL'
        if order_type == 'MKT':
            ib_order = ibi.order.MarketOrder(buy_sell, abs(order_amount))
            trade_object = self._broker_handle.placeOrder(contract, ib_order)
            return trade_object
        else:
            # TBD: Support other order types
            raise ValueError('IBOrderManager.order_wrapper: Order type ' + order_type + ' is supported.')

    def place_order(self,
                    strategy_name: str,
                    combo_name: str,
                    time_stamp: str,
                    contract_index: int,
                    contract: IBBrokerAPI.IBBrokerContract,
                    unit: float
                    ) -> None:
        """
        Place order that is requested by strategy.

        Parameters
        ----------
        strategy_name: str
            Name of the strategy placing the order.
        combo_name: str
            Name of the combo. Each combo in a strategy should have a unique name.
        time_stamp: str
            Timestamp.
        contract_index: int
            Index of the contract as found in contract array.
        contract: IBBrokerAPI.IBBrokerContract
            Contract to be traded.
        unit: float
            Number of unit to trade.
        """
        order_notional = unit * self._combo_weight[(strategy_name, combo_name)][contract_index]
        if abs(order_notional) > utils.EPSILON:
            # Update order status
            trade_object = self.order_wrapper(contract.get_contract(), order_notional)
            trade_object.statusEvent += self.update_order_status

            self.order_info_df = self.order_info_df.append({'strategy_name': strategy_name,
                                                            'combo_name': combo_name,
                                                            'contract_index': contract_index,
                                                            'combo_unit': unit,
                                                            'dangling_order': order_notional,
                                                            'order_id': trade_object.orderStatus.orderId,
                                                            'time_stamp': time_stamp},
                                                           ignore_index=True)

    def update_order_status(self,
                            trade: ibi.order.Trade) -> None:
        """
        Update order status.

        Parameters
        ----------
        trade: ibi.order.Trade
            IB Trade object.
        """
        order_id = trade.orderStatus.orderId
        average_fill_price = trade.orderStatus.avgFillPrice
        filled_amount = trade.orderStatus.filled

        if abs(filled_amount) > utils.EPSILON:
            # Register leg with PortfolioManager
            if order_id not in self.order_info_df['order_id'].to_list():
                # Probably receiving order status update after it is already completely filled; do nothing
                return
            combo_name = self.order_info_df.loc[self.order_info_df['order_id'] == order_id, 'combo_name'].iloc[0]
            time_stamp = self.order_info_df.loc[self.order_info_df['order_id'] == order_id, 'time_stamp'].iloc[0]
            if sum(abs(self.order_info_df.loc[(self.order_info_df['combo_name'] == combo_name) &
                                              (self.order_info_df['time_stamp'] == time_stamp), 'dangling_order'])) \
                    < utils.EPSILON:
                # All legs in the combo already filled.
                return

            # Update dangling order status.
            order_direction = 1.0 if trade.order.action == 'BUY' else -1.0
            self.order_info_df.loc[self.order_info_df['order_id'] == order_id, 'dangling_order']\
                -= order_direction * filled_amount

            # Record leg entry price
            self.order_info_df.loc[self.order_info_df['order_id'] == order_id, 'entry_price'] = average_fill_price

            # Communicate update to portfolio manager.
            self._portfolio_manager.register_contract_trade(symbol=trade.contract.symbol,
                                                            type_=trade.contract.secType,
                                                            exchange=trade.contract.exchange,
                                                            currency=trade.contract.currency,
                                                            position=filled_amount)

            # Check if all legs in the combo are filled, if so update portfolio manager accordingly.
            if sum(abs(self.order_info_df.loc[(self.order_info_df['combo_name'] == combo_name) &
                                              (self.order_info_df['time_stamp'] == time_stamp), 'dangling_order'])) \
                    < utils.EPSILON:
                # Calculate combo entry price.
                leg_entry_prices = (self.order_info_df.loc[(self.order_info_df['combo_name'] == combo_name) &
                                    (self.order_info_df['time_stamp'] == time_stamp),
                                                           ['contract_index', 'entry_price']])
                leg_entry_prices.sort_values(['contract_index'], ascending=True, inplace=True)
                strategy_name = self.order_info_df.loc[
                    self.order_info_df['combo_name'] == combo_name, 'strategy_name'].iloc[0]
                combo_def = [self._combo_weight[(strategy_name, combo_name)][index]
                             for index in leg_entry_prices['contract_index']]
                combo_entry_price = sum([x * y for x, y in zip(leg_entry_prices['entry_price'].tolist(), combo_def)])

                # Register with portfolio manager
                combo_unit = (self.order_info_df.loc[(self.order_info_df['combo_name'] == combo_name) &
                                                     (self.order_info_df['time_stamp'] == time_stamp),
                              'combo_unit'].iloc[0])
                self._portfolio_manager.register_combo_trade(
                    strategy_name=strategy_name,
                    combo_name=combo_name,
                    combo_unit=combo_unit,
                    combo_def=combo_def,
                    combo_entry_price=combo_entry_price)

                # Clean up dataframe by deleting rows already filled
                self.order_info_df = self.order_info_df[~(self.order_info_df['combo_name'] == combo_name) |
                                                        ~(self.order_info_df['time_stamp'] == time_stamp)]



from datetime import datetime
import pandas as pd
from typing import List


class TimeDouble:
    """
    Basic timestamped data class.
    """

    def __init__(self,
                 data_name: str,
                 timestamp: datetime,
                 value: float) -> None:
        """
        Initialize time double.

        Parameters
        ----------
        data_name: str
            Name of the data represented.
        timestamp: datetime
            Time of the realization of the data point.
        value: float
            Numerical value of the data.
        """
        self._data_name = data_name
        self._timestamp = timestamp
        self._value = value

    def get_name(self) -> str:
        return self._data_name

    def get_time(self) -> datetime:
        return self._timestamp

    def get_value(self) -> float:
        return self._value


class PriceTimeSeries:
    """
    Class for price time series.
    """
    def __init__(self,
                 name: str,
                 timestamp: List[datetime],
                 price: List[float]) -> None:
        """
        Initialize price time series.

        Parameters
        ----------
        name: str
            Name of the data represented.
        timestamp: List[datetime]
            Time of the realization of the data point.
        price: List[float]
            Numerical value of the data.
        """
        self._name = name
        self._timestamp = timestamp
        self._price = price

    def get_price(self) -> List[float]:
        return self._price

    def get_timestamp(self) -> List[datetime]:
        return self._timestamp

    def get_name(self) -> str:
        return self._name


class DataUtils:
    @staticmethod
    def aggregate_yahoo_data(ticker_name_list: List[str],
                             path: str) -> List[PriceTimeSeries]:
        """
        Aggregate multiple ticker data (as Yahoo csv file).

        Parameters
        ----------
        ticker_name_list: List[str]
            A list of ticker names that are expected in the folder.
        path: str
            Path to files.

        Returns
        -------
            List[PriceTimeSeries]
                Aggregated data.
        """
        # TBD: Ideally this function should also perform timestamp matching, filling forward, NaN handling etc.
        output: List[PriceTimeSeries] = []

        # Read one by one
        for ticker in ticker_name_list:
            yahoo_data: PriceTimeSeries = DataUtils.read_yahoo_data(ticker, path)
            output.append(yahoo_data)
        return output

    @staticmethod
    def read_yahoo_data(ticker: str,
                        path: str) -> PriceTimeSeries:
        """
        Read Yahoo Finance csv file.

        Parameters
        ----------
        ticker: str
            Ticker name that are expected in the folder.
        path: str
            Path to files.

        Returns
        -------
            PriceTimeSeries
                Data read.
        """
        file_extension = '.csv'
        df = pd.read_csv(path + ticker + file_extension)
        if str(df['Date'][0]).find('/') != -1:
            yahoo_date_format = '%m/%d/%Y'
        elif str(df['Date'][0]).find('-') != -1:
            yahoo_date_format = '%Y-%m-%d'
        else:
            raise ValueError('DataUtils.read_yahoo_data: Unknown date format.')
        datetime_list = [datetime.strptime(x, yahoo_date_format) for x in df['Date'].tolist()]
        return PriceTimeSeries(ticker, datetime_list, df['Adj Close'].tolist())



from abc import ABC, abstractmethod
from collections import deque
from datetime import timedelta
from datetime import datetime
import enum
from eventkit import Event
import math
import pandas as pd
from typing import List, Deque, Dict

from . import brokerinterface
from . import datacontainer
from . import utils


class WeightingScheme(enum.Enum):
    """
    Enum class for portfolio weighting scheme
    """
    Equal = 1


class TradeCombos:
    """
    Class for trade combos, which is a basket of contracts traded by a strategy.
    """

    def __init__(self,
                 contract_array: List[Event],
                 combo_definition: Dict[str, List[float]] = None) -> None:
        """
        Initialize trade combo.

        Parameters
        ----------
        contract_array: List[Event]
            List of event subscriptions on contracts to be traded.
        combo_definition: Dict[str, List[float]]
            A dictionary, with keys being combo names and values being weights to be traded.
        """
        if combo_definition is not None:
            for key, value in combo_definition.items():
                if len(contract_array) != len(value):
                    raise ValueError('TradeCombo.__init__: Different numbers of contracts and weights.')
            self._combo_definition = combo_definition
        else:
            self._combo_definition = None
        self._contract_array = contract_array

    def get_combo_def(self) -> Dict[str, List[float]]:
        if self._combo_definition is not None:
            return self._combo_definition
        else:
            return None

    def get_contract_array(self) -> List[Event]:
        return self._contract_array


class StrategyBase(ABC):
    """
    Base class for strategy.
    """
    CASH_ACCOUNT_NAME = 'cash_account'

    _strategy_classes_registry = {}

    _strategy_signature = None

    @classmethod
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if cls._strategy_signature is None:
            raise KeyError('StrategyBase: Missing signature for ' + str(cls))
        elif cls._strategy_signature in cls._strategy_classes_registry:
            raise KeyError('StrategyBase: Conflict in signature ' + cls._strategy_signature)
        else:
            cls._strategy_classes_registry[cls._strategy_signature] = cls

    @classmethod
    def get_class_registry(cls):
        return cls._strategy_classes_registry

    def __new__(cls,
                signature_str: str,
                input_signal_array: List[Event],
                trade_combo: TradeCombos,
                warmup_length: int,
                initial_capital: float = 100.0,
                order_manager: brokerinterface.OrderManagerBase = None,
                **kwargs):
        """
        Create derived strategy object.

        Parameters
        ----------
        signature_str: str
            Unique signature of the strategy.
        input_signal_array: List[Event]
            List of event subscription that is required for trade decision.
        trade_combo: TradeCombos
            TradeCombos object that defines the combos to be traded. Note that if combo definition is missing,
            all trades from this strategy will be under a generic combo.
        warmup_length: int
            Number of data points to 'burn'.
        initial_capital: float
            Initial capital.
        order_manager: OrderManager
            OrderManager object for live trading.
        """
        if signature_str not in cls.get_class_registry():
            raise ValueError('StrategyBase.__new__: ' + signature_str + ' is not a valid key.')

        my_strategy_obj = super().__new__(cls.get_class_registry()[signature_str])

        # TBD: Use pandas df to store some of the member fields
        my_strategy_obj._warmup_length = warmup_length
        my_strategy_obj._mtm = initial_capital
        if order_manager is None:
            my_strategy_obj._is_live_trading = False
        else:
            my_strategy_obj._is_live_trading = True
            my_strategy_obj._order_manager = order_manager

        my_strategy_obj._strategy_active = True  # A strategy is active by default
        my_strategy_obj._initialize_signal_time_storage(input_signal_array)
        contract_array = trade_combo.get_contract_array()
        my_strategy_obj._contract_array = contract_array
        if not trade_combo.get_combo_def() is None:
            my_strategy_obj._combo_def = trade_combo.get_combo_def()
        else:
            my_strategy_obj._combo_def = None
        my_strategy_obj._initialize_contract_time_storage(contract_array)
        my_strategy_obj._input_signal_array = input_signal_array

        position_temp = {security.name(): 0.0 for security in contract_array}
        # Remember to add cash account
        position_temp[my_strategy_obj.CASH_ACCOUNT_NAME] = initial_capital
        my_strategy_obj._contract_positions = position_temp

        my_strategy_obj._combo_positions = {combo_name: 0.0 for combo_name in my_strategy_obj._combo_def.keys()}

        my_strategy_obj._combo_order = {combo_name: 0.0 for combo_name in my_strategy_obj._combo_def.keys()}
        my_strategy_obj._average_entry_price = {security.name(): 0.0 for security in contract_array}
        my_strategy_obj._mtm_price = {security.name(): 0.0 for security in contract_array}

        my_strategy_obj._mtm_history = [initial_capital]

        return my_strategy_obj

    def __init__(self,
                 signature_str: str,
                 input_signal_array: List[Event],
                 trade_combo: TradeCombos,
                 warmup_length: int,
                 initial_capital: float = 100.0,
                 order_manager: brokerinterface.OrderManagerBase = None,
                 **kwargs) -> None:
        pass

    def _initialize_signal_time_storage(self, input_signal_array: List[Event]) -> None:
        """
        Initialize storage.

        Parameters
        ----------
        input_signal_array: List[Event]
            List of signal subscription that is required for ordering.
        """
        signal_storage = {}
        signal_time_storage = {}
        signal_slot_storage = []
        for input_signal in input_signal_array:
            signal_storage[input_signal.name()] = deque([])
            signal_time_storage[input_signal.name()] = deque([])

            this_signal_slot = SignalSlot(input_signal.name(), [self])
            signal_slot_storage.append(this_signal_slot)
            # Connect event to slot
            input_signal += this_signal_slot.on_event
        self._signal_storage = signal_storage
        self._signal_time_storage = signal_time_storage
        self._signal_slot_storage = signal_slot_storage

    def _initialize_contract_time_storage(self, contract_array: List[Event]) -> None:
        """
        Initialize storage.

        Parameters
        ----------
        contract_array: List[Event]
            List of contracts.
        """
        contract_time_storage = {}
        mtm_price_slot_storage = []
        for input_event in contract_array:
            contract_time_storage[input_event.name()] = deque([])

            this_mtm_price_slot = MTMPriceSlot(input_event.name(), [self])
            mtm_price_slot_storage.append(this_mtm_price_slot)
            # Connect event to slot
            input_event += this_mtm_price_slot.on_event
        self._contract_time_storage = contract_time_storage
        self._mtm_price_slot_storage = mtm_price_slot_storage

    def get_strategy_active(self) -> bool:
        return self._strategy_active

    def get_strategy_name(self) -> str:
        return self._strategy_name

    def set_strategy_active(self, activity: bool) -> None:
        self._strategy_active = activity

    def check_all_signals_received(self, signal_name: str) -> bool:
        output = True
        num_signal = len(self._input_signal_array)
        if num_signal == 1:
            return output  # If there is only one incoming signal stream, no need to check
        for i in range(num_signal):
            event_i = self._input_signal_array[i]
            if len(self.get_signal_time_by_name(event_i.name())) == 0:
                return False

            time_diff = self.get_signal_time_by_name(signal_name)[-1] - self.get_signal_time_by_name(event_i.name())[-1]
            if time_diff > timedelta(microseconds=1):
                return False
        return output

    def update_data(self, new_data: datacontainer.TimeDouble) -> None:
        """
        Update signal storage.

        Parameters
        ----------
        new_data: TimeDouble
            Incoming new data.
        """
        # Extend storage
        self._signal_storage[new_data.get_name()].append(new_data.get_value())
        self._signal_time_storage[new_data.get_name()].append(new_data.get_time())
        # Remove oldest if warming up is complete
        if len(self._signal_storage[new_data.get_name()]) > self._warmup_length:
            self._signal_storage[new_data.get_name()].popleft()
            self._signal_time_storage[new_data.get_name()].popleft()

    def get_signal_time_by_name(self, signal_name: str) -> Deque:
        return self._signal_time_storage[signal_name]

    def get_signal_by_name(self, signal_name: str) -> Deque:
        return self._signal_storage[signal_name]

    def check_all_contract_data_received(self, data_name: str) -> bool:
        output = True
        num_data = len(self._contract_array)
        if num_data == 1:
            return output  # If there is only one incoming data stream, no need to check
        for i in range(num_data):
            event_name_i = self._contract_array[i].name()
            if len(self.get_contract_time_by_name(event_name_i)) == 0:
                return False

            time_diff = self.get_contract_time_by_name(data_name)[-1] - self.get_contract_time_by_name(event_name_i)[-1]
            if time_diff > timedelta(microseconds=1):
                return False
        return output

    def update_mtm(self, new_data: datacontainer.TimeDouble) -> None:
        """
        Update data storage used for MTM calculation.

        Parameters
        ----------
        new_data: TimeDouble
            Incoming new data.
        """
        self._mtm_price[new_data.get_name()] = new_data.get_value()
        self.get_contract_time_by_name(new_data.get_name()).append(new_data.get_time())
        if len(self.get_contract_time_by_name(new_data.get_name())) > 1:
            self.get_contract_time_by_name(new_data.get_name()).popleft()
        # Update PM combo mtm price
        if self._is_live_trading:
            # Build a dataframe
            combo_mtm = pd.DataFrame(columns=['strategy_name', 'combo_name', 'combo_mtm_price'])
            for combo_name in self._combo_def.keys():
                combo_mtm = combo_mtm.append({'strategy_name': self._strategy_name,
                                              'combo_name': combo_name,
                                              'combo_mtm_price': self.get_combo_mtm_price(combo_name)},
                                             ignore_index=True)
            # Update
            self._order_manager.get_portfolio_manager().update_combo_mtm_price(combo_mtm)
        pass

    def get_mtm_history(self) -> List[float]:
        return self._mtm_history

    def get_mtm_price_by_name(self, data_name: str) -> float:
        """Return the MTM price of a contract."""
        return self._mtm_price[data_name]

    def get_combo_mtm_price(self, combo_name: str) -> float:
        """
        Retrieve combo mtm price.


        Parameters
        ----------
        combo_name: str
            Combo name.
        """
        combo_def = self._combo_def[combo_name]
        return sum(i[0] * i[1] for i in zip(combo_def, self._mtm_price.values()))

    def get_contract_time_by_name(self, data_name: str) -> Deque:
        return self._contract_time_storage[data_name]

    def get_signal_containing_string(self,
                                     input_string: str) -> Deque:
        """
        Retrieve the signals that contain a certain string in the name.

        Parameters
        ----------
        input_string: str
            String that is supposed to be part of the name.

        Returns
        -------
            Deque
                The signal data that is stored.
        """
        for key, value in self._signal_storage.items():
            if input_string in key:
                return value
        # If not found
        raise ValueError('StrategyBase.get_signal_containing_string: Signal not found.')

    def send_order(self) -> None:
        """
        Send pending orders.
        """
        if self._is_live_trading:
            self.send_order_live()
        else:
            self.send_order_backtest()

    def send_order_backtest(self) -> None:
        """
        Send pending orders and use latest price to calculate average cost.
        """
        if self._combo_order is None:
            return
        for combo_name, amount in self._combo_order.items():
            if combo_name in self._combo_def:
                combo_def = self._combo_def[combo_name]
                self._combo_positions[combo_name] += amount
                for weight_i, contract_i in zip(combo_def, self._contract_array):
                    # Adjust cash account
                    name_i = contract_i.name()
                    self._contract_positions[self.CASH_ACCOUNT_NAME] -= amount * weight_i * self._mtm_price[name_i]
                    # Calculate average cost
                    self._average_entry_price[name_i] = self._calculate_new_average_entry_price(
                        self._average_entry_price[name_i],
                        self._mtm_price[name_i],
                        self._contract_positions[name_i],
                        amount * weight_i)
                    # Modify current position
                    self._contract_positions[name_i] += amount * weight_i
            else:
                # Some strategies do not have predefined combos;
                # for them we simply place orders for individual contract.
                contract_name = combo_name
                # Adjust cash account
                self._contract_positions[self.CASH_ACCOUNT_NAME] -= amount * self._mtm_price[contract_name]
                # Calculate average cost
                self._average_entry_price[contract_name] = self._calculate_new_average_entry_price(
                    self._average_entry_price[contract_name],
                    self._mtm_price[contract_name],
                    self._contract_positions[contract_name],
                    amount)
                # Modify current position
                self._contract_positions[contract_name] += amount
            # Reset pending order
            self._combo_order[combo_name] = 0.0

    def send_order_live(self) -> None:
        """
        Send pending orders to broker.
        """
        if self._combo_order is None:
            return
        contract_array = [self._order_manager.get_event_contract_dict()[i.name()] for i in self._contract_array]
        for combo_name, amount in self._combo_order.items():
            if combo_name in self._combo_def:
                combo_def = self._combo_def[combo_name]
                self._order_manager.register_combo_level_info(self._strategy_name,
                                                              contract_array,
                                                              combo_def,
                                                              combo_name)
                time_stamp = str(datetime.now())
                for i in range(len(combo_def)):
                    self._order_manager.place_order(self._strategy_name,
                                                    combo_name,
                                                    time_stamp,
                                                    i,
                                                    contract_array[i],
                                                    amount
                                                    )
                self._combo_positions[combo_name] += amount
            else:
                # Some strategies do not have predefined combos;
                # for them we simply place orders for individual contract.
                raise ValueError('StrategyBase.send_order_live: Non-combo ordering not implemented yet.')

    def _calculate_new_average_entry_price(self,
                                           old_average_entry_price: float,
                                           new_entry_price: float,
                                           existing_position: float,
                                           new_pending_order: float) -> float:
        """
        Calculate new average entry price.

        Parameters
        ----------
        old_average_entry_price: float
            Existing average entry price.
        new_entry_price: float
            New entry price.
        existing_position: float
            Existing position.
        new_pending_order: float
            New pending order.

        Returns
        -------
            float
                The new average entry price.
        """
        if math.copysign(1, existing_position) != math.copysign(1, existing_position + new_pending_order):
            # If going from net long to net short or vice versa
            return new_entry_price
        elif abs(existing_position + new_pending_order) > 0.0:
            return (existing_position * old_average_entry_price + new_pending_order * new_entry_price) / (
                existing_position + new_pending_order)
        else:
            return new_entry_price

    def calculate_mtm(self) -> None:
        """
        Mark to market.
        """
        mtm = 0.0
        for key, value in self._mtm_price.items():
            mtm += value * self._contract_positions[key]
        mtm += self._contract_positions[self.CASH_ACCOUNT_NAME]
        self._update_mtm_history(mtm)

    def _update_mtm_history(self, new_mtm: float) -> None:
        """
        Update MTM history.

        Parameters
        ----------
        new_mtm: float
            Latest MTM.
        """
        self._mtm_history.append(new_mtm)

    def has_pending_order(self) -> bool:
        """
        Check if there is any pending order.
        """
        for key, value in self._combo_order.items():
            if abs(value) > 0.0:
                return True
        return False

    def has_specific_pending_order(self, combo_name: str) -> bool:
        """
        Check if a specific contract has any pending order.
        """
        if abs(self._combo_order[combo_name]) > 0.0:
            return True
        else:
            return False

    def is_currently_long(self, combo_name: str) -> bool:
        return self._combo_positions[combo_name] > 0.0

    def is_currently_short(self, combo_name: str) -> bool:
        return self._combo_positions[combo_name] <= 0.0

    def is_currently_flat(self, combo_name: str) -> bool:
        return abs(self._combo_positions[combo_name]) <= utils.EPSILON

    def get_warmup_length(self) -> int:
        return self._warmup_length

    def get_is_live_trading(self) -> bool:
        return self._is_live_trading

    @abstractmethod
    def make_order_decision(self) -> Dict[str, float]:
        """
        Virtual method to be implemented by the derived class.

        Returns
        -------
            Dict[str, float]
                A dictionary of combo name to trade, and amount to trade.
        """
        pass


class MACrossingStrategy(StrategyBase):
    """
    MA Crossing Strategy
    """

    _strategy_signature = 'MACrossing'

    def __init__(self,
                 signature_str: str,
                 input_signal_array: List[Event],
                 trade_combo: TradeCombos,
                 warmup_length: int,
                 initial_capital: float = 100.0,
                 order_manager: brokerinterface.OrderManagerBase = None) -> None:
        """
        Initialize MA Crossing Strategy.

        Parameters
        ----------
        signature_str: str
            Unique signature of the strategy.
        input_signal_array: List[Event]
            List of event subscription that is required for trade decision.
        trade_combo: TradeCombos
            TradeCombos object that defines the combos to be traded. Note that if combo definition is missing,
            all trades from this strategy will be under a generic combo.
        warmup_length: int
            Number of data points to 'burn'.
        initial_capital: float
            Initial capital.
        order_manager: OrderManager
            OrderManager object for live trading.
        """
        if len(trade_combo.get_combo_def()) != 1:
            raise ValueError('MACrossingStrategy.__init__: Too many combos specified.')
        self._strategy_name = input_signal_array[0].name() + self._strategy_signature
        self._long_signal_name, self._short_signal_name = self._deduce_signal_name(input_signal_array)

    def _deduce_signal_name(self, input_signal_array: List[Event]) -> (str, str):
        """
        Deduce the name of the signal with longer (shorter) average window.

        Parameters
        ----------
        input_signal_array: List[Event]
            List of event subscription that is required for trade decision.

        Returns
        -------
            str, str
                The long and short window signal names respectively.
        """
        name_1 = input_signal_array[0].name()
        name_2 = input_signal_array[1].name()
        delimiter = '_'
        substr_1 = name_1.split(delimiter)[-1]
        substr_2 = name_2.split(delimiter)[-1]
        if not (isinstance(int(substr_1), int) and isinstance(int(substr_2), int)):
            raise TypeError('MACrossingStrategy._deduce_signal_name: Signal name in wrong format.')
        if int(substr_1) > int(substr_2):
            return name_1, name_2
        else:
            return name_2, name_1

    def make_order_decision(self) -> Dict[str, float]:
        combo_name = next(iter(self._combo_def))
        current_price = self.get_combo_mtm_price(combo_name)
        if not self.has_specific_pending_order(combo_name):
            if self._signal_storage[self._long_signal_name][-1] >= self._signal_storage[self._short_signal_name][-1] \
                    and (self.is_currently_long(combo_name) or self.is_currently_flat(combo_name)):
                return {combo_name: -self._mtm_history[-1] / current_price - self._combo_positions[combo_name]}
            elif self._signal_storage[self._long_signal_name][-1] < self._signal_storage[self._short_signal_name][-1] \
                    and (self.is_currently_short(combo_name) or self.is_currently_flat(combo_name)):
                return {combo_name: self._mtm_history[-1] / current_price - self._combo_positions[combo_name]}
            else:
                return {combo_name: 0.0}
        else:
            return {combo_name: 0.0}


class VAAStrategy(StrategyBase):
    """
    VAAStrategy
    https://indexswingtrader.blogspot.com/2017/07/breadth-momentum-and-vigilant-asset.html
    Given a top selection T and a breadth protection threshold B, for each month:
    - Compute 13612W momentum for each asset
    - Pick the best performing assets in the "risk-on" universe as top T
    - Pick the best asset in the "risk-off" universe as safety asset for "cash"
    - Compute the number of assets with non-positive momentum in the "risk-on" universe (b)
    - Compute b/B and round down to multiples of 1/T as "cash fraction" CF for "easy trading"
    - Replace CF of top T by "cash" asset as selected in step 3
    """

    _strategy_signature = 'VAA'

    def __init__(self,
                 signature_str: str,
                 input_signal_array: List[Event],
                 trade_combo: TradeCombos,
                 warmup_length: int,
                 initial_capital: float = 100.0,
                 order_manager: brokerinterface.OrderManagerBase = None,
                 **kwargs) -> None:
        """
        Initialize VAA rotation Strategy.

        Parameters
        ----------
        signature_str: str
            Unique signature of the strategy.
        input_signal_array: List[Event]
            List of event subscription that is required for trade decision.
        trade_combo: TradeCombos
            TradeCombos object that defines the combos to be traded. Note that if combo definition is missing,
            all trades from this strategy will be under a generic combo.
        warmup_length: int
            Number of data points to 'burn'.
        initial_capital: float
            Initial capital.
        order_manager: OrderManager
            OrderManager object for live trading.

        Other Parameters
        ----------------
        risk_on_size: int
            Number of risk-on assets within the input signals.
        num_assets_to_hold: int
            Portfolio size in terms of number of holding.
        breadth_protection_threshold: int
            Threshold for protection.
        weighting_scheme: WeightingScheme
            Portfolio weighting scheme.
        """
        if len(trade_combo.get_contract_array()) < kwargs['num_assets_to_hold']:
            raise ValueError('VAAStrategy.__init__: num_assets_to_hold must be smaller than universe size.')
        if len(input_signal_array) != len(trade_combo.get_contract_array()):
            raise ValueError('VAAStrategy.__init__: input_signal_array and trade combo contracts '
                             'should be of the same size.')
        all_input_names = [x.name() for x in input_signal_array]
        self._strategy_name = ''.join(all_input_names) + "_" + self._strategy_signature
        self._num_assets_to_hold = kwargs['num_assets_to_hold']
        self._weighting_scheme = kwargs['weighting_scheme']
        self._breadth_protection_threshold = kwargs['breadth_protection_threshold']
        self._risk_on_names, self._risk_off_names = self.initialize_names(input_signal_array, kwargs['risk_on_size'])

    def initialize_names(self,
                         input_signal_array: List[Event],
                         risk_on_size: int) -> (List[str], List[str]):
        """
        Initialize the risk-on and risk-off names.

        Parameters
        ----------
        input_signal_array: List[Event]
            List of event subscription that is required for trade decision.
        risk_on_size: int
            Number of risk-on assets within the input signals.

        Returns
        -------
            List[str], List[str]
                The risk-on and risk-off name lists, respectively.
        """
        output_on = []
        output_off = []
        for i in range(len(input_signal_array)):
            output_on.append(input_signal_array[i].name()) if i < risk_on_size else \
                output_off.append(input_signal_array[i].name())
        return output_on, output_off

    def get_risk_on_off_signals(self, is_on: bool) -> Dict[str, Deque[float]]:
        """
        Retrieve risk-on or risk-off signals.

        Parameters
        ----------
        is_on: bool
            Whether we are interested in risk-on or risk-off.

        Returns
        -------
            Dict[str, Deque[float]]
                A dictionary containing signals on the selection of names.
        """
        output = {}
        names = self._risk_on_names if is_on else self._risk_off_names
        for name in names:
            output[name] = self.get_signal_by_name(name)
        return output

    def make_order_decision(self) -> Dict[str, float]:
        # Identify top performing assets from the risk-on assets
        on_signals = self.get_risk_on_off_signals(True)
        top_risk_on_performers = sorted(
            on_signals.items(), key=lambda e: e[1][-1], reverse=True)[:self._num_assets_to_hold]
        # Identify top performing assets from the risk-off assets
        off_signals = self.get_risk_on_off_signals(False)
        top_risk_off_performers = sorted(
            off_signals.items(), key=lambda e: e[1][-1], reverse=True)[0]
        # Compute the number of assets with non-positive momentum in the "risk-on" universe
        max_signals = map(max, on_signals.values())
        num_non_positive = sum(i < 0.0 for i in max_signals)

        rounded_cash_fraction = max(0.0, min(1.0, (1.0 / float(self._num_assets_to_hold))
                                             * math.floor(float(num_non_positive) * float(self._num_assets_to_hold)
                                                          / float(self._breadth_protection_threshold))))

        # Rebalance - Buy new holdings
        output = self.calculate_weight(top_risk_on_performers, top_risk_off_performers, rounded_cash_fraction)

        # Rebalance - Sell old holdings
        for pos_k, pos_v in self._contract_positions.items():
            if (abs(pos_v) > 0.0) and (pos_k != self.CASH_ACCOUNT_NAME):
                if pos_k not in output:
                    output[pos_k] = -pos_v
                else:
                    output[pos_k] -= pos_v
        return output

    def calculate_weight(self,
                         top_performers: List[str],
                         top_risk_off_performer: str,
                         rounded_cash_fraction: float) -> Dict[str, float]:
        """
        Calculate the weight to trade.

        Parameters
        ----------
        top_performers: List[str]
            Top risk-on performers' performances.
        top_risk_off_performer: str
            Top risk-off performer's performance.
        rounded_cash_fraction: float
            Rounded fraction to be held in cash.

        Returns
        -------
            Dict[str, float]
                Trade decision.
        """
        num_risk_on_to_buy = round(1.0 - rounded_cash_fraction) / (1.0 / float(self._num_assets_to_hold))
        output = {}
        num_risk_on_bought = 0
        for t in top_performers:
            if num_risk_on_bought < num_risk_on_to_buy:
                contract_name = t[0].split('_')[0]
                if self._weighting_scheme == WeightingScheme.Equal:
                    asset_weight = 1.0 / self._num_assets_to_hold
                else:
                    raise ValueError('VAAStrategy.calculate_weight: unknown weighting scheme '
                                     + str(self._weighting_scheme))
                output[contract_name] = asset_weight
                num_risk_on_bought += 1
            else:
                contract_name = top_risk_off_performer[0].split('_')[0]
                output[contract_name] = rounded_cash_fraction\
                    * self.get_mtm_history()[-1] / self.get_mtm_price_by_name(contract_name)
                break
        return output


class SlotBase:
    """
    Base class for slot, to be derived for Strategy.
    """

    def __init__(self,
                 data_name: str,
                 parent_strategies: List[StrategyBase]) -> None:
        """
        Initialize base slot.

        Parameters
        ----------
        data_name: str
            Name of the input data.
        parent_strategies: List[StrategyBase]
            Strategies that listens to this data.
        """
        self._data_name = data_name
        self._parent_strategies = parent_strategies

    @abstractmethod
    def on_event(self, new_data: datacontainer.TimeDouble) -> None:
        """
        Perform the action upon getting an event.
        """
        pass


class SignalSlot(SlotBase):
    """
    Class for signal slot.
    """

    def __init__(self,
                 data_name: str,
                 parent_strategies: List[StrategyBase]) -> None:
        super().__init__(data_name, parent_strategies)

    def on_event(self, new_signal: datacontainer.TimeDouble) -> None:
        """
        Perform the action upon getting an event.

        When there is an event (arrival of signal) we want to
        - Update signal storage
        - Make trade decision if warming up is complete
        - Send order

        Parameters
        ----------
        new_signal: TimeDouble
            Incoming new signal.
        """
        for _parent in self._parent_strategies:
            # 1. Update signal storage
            _parent.update_data(new_signal)

            # 2. If warming up is complete and all data arrived, make trade decision
            if len(_parent.get_signal_by_name(new_signal.get_name())) == _parent.get_warmup_length()\
                    and _parent.check_all_signals_received(new_signal.get_name()):
                if _parent.get_strategy_active():
                    _parent._combo_order = _parent.make_order_decision()
                    # 3. Send order
                    if _parent.get_is_live_trading():
                        _parent.send_order()


class MTMPriceSlot(SlotBase):
    """
    Class for marking to market price slot.
    """

    def __init__(self,
                 data_name: str,
                 parent_strategies: List[StrategyBase]) -> None:
        super().__init__(data_name, parent_strategies)

    def on_event(self, new_data: datacontainer.TimeDouble) -> None:
        """
        Perform the action upon getting an event.

        When there is an event (arrival of contract prices) we want to
        - Update mtm price storage
        - Mark to market

        Parameters
        ----------
        new_data: TimeDouble
            Incoming new data.
        """
        for _parent in self._parent_strategies:
            # 1. Update mtm price storage
            _parent.update_mtm(new_data)
            if _parent.check_all_contract_data_received(new_data.get_name()):
                # 2. Send order
                if not _parent.get_is_live_trading():
                    _parent.send_order()
                # 3. Mark to market
                _parent.calculate_mtm()



from abc import ABC, abstractmethod
from collections import deque
from datetime import timedelta
from eventkit import Event
from typing import List, Deque

from . import datacontainer


class SignalBase(ABC):
    """
    Base class for signal.
    """

    _signal_classes_registry = {}

    _signal_signature = None

    @classmethod
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if cls._signal_signature is None:
            raise KeyError('SignalBase: Missing signature for ' + str(cls))
        elif cls._signal_signature in cls._signal_classes_registry:
            raise KeyError('SignalBase: Conflict in signature ' + cls._signal_signature)
        else:
            cls._signal_classes_registry[cls._signal_signature] = cls

    @classmethod
    def get_class_registry(cls):
        return cls._signal_classes_registry

    def __new__(cls,
                signature_str: str,
                input_data_array: List[Event],
                warmup_length: int,
                **kwargs):
        """
        Create derived signal object.

        Parameters
        ----------
        signature_str: str
            Unique signature of the signal.
        input_data_array: List[Event]
            List of event subscription that is required for signal calculation.
        warmup_length: int
            Number of data points to 'burn'.
        """
        if signature_str not in cls.get_class_registry():
            raise ValueError('SignalBase.__new__: ' + signature_str + ' is not a valid key.')

        my_signal_obj = super().__new__(cls.get_class_registry()[signature_str])

        # Initialization
        my_signal_obj._input_data_array = input_data_array
        my_signal_obj._warmup_length = warmup_length
        my_signal_obj._initialize_data_time_storage(input_data_array)
        my_signal_obj._signature_str = signature_str

        return my_signal_obj

    def __init__(self,
                 signature_str: str,
                 input_data_array: List[Event],
                 warmup_length: int,
                 **kwargs) -> None:
        pass

    def _initialize_data_time_storage(self, input_data_array: List[Event]) -> None:
        """
        Initialize storage.

        Parameters
        ----------
        input_data_array: List[Event]
            List of event subscription that is required for signal calculation.
        """
        data_storage = {}
        time_storage = {}
        data_slot_storage = []
        for input_data in input_data_array:
            data_storage[input_data.name()] = deque([])
            time_storage[input_data.name()] = deque([])

            this_slot = DataSlot(input_data.name(), [self])
            data_slot_storage.append(this_slot)
            # Connect event to slot
            input_data += this_slot.on_event
        self._data_storage = data_storage
        self._time_storage = time_storage
        self._data_slot_storage = data_slot_storage

    def check_all_received(self, data_name: str) -> bool:
        output = True
        num_data = len(self._input_data_array)
        if num_data == 1:
            return output  # If there is only one incoming data stream, no need to check
        for i in range(num_data):
            event_i = self._input_data_array[i]
            if len(self.get_time_by_name(event_i.name())) == 0:
                return False

            time_diff = self.get_time_by_name(data_name)[-1] - self.get_time_by_name(event_i.name())[-1]
            if time_diff > timedelta(microseconds=1):
                return False
        return output

    def update_data(self, new_data: datacontainer.TimeDouble) -> None:
        """
        Update data storage.

        Parameters
        ----------
        new_data: TimeDouble
            Incoming new data.
        """
        # Extend storage
        self._data_storage[new_data.get_name()].append(new_data.get_value())
        self._time_storage[new_data.get_name()].append(new_data.get_time())
        # Remove oldest if warming up is complete
        if len(self._data_storage[new_data.get_name()]) > self._warmup_length:
            self._data_storage[new_data.get_name()].popleft()
            self._time_storage[new_data.get_name()].popleft()

    def get_data_by_name(self, data_name: str) -> Deque:
        return self._data_storage[data_name]

    def get_time_by_name(self, data_name: str) -> Deque:
        return self._time_storage[data_name]

    def get_warmup_length(self) -> int:
        return self._warmup_length

    def get_signal_name(self) -> str:
        return self._signal_name

    def get_signal_event(self) -> Event:
        return self._signal_event

    @abstractmethod
    def calculate_signal(self) -> float:
        """
        Virtual method to be implemented by the derived class.
        """
        pass


class MASignal(SignalBase):
    """
    Moving average signal.
    """

    _signal_signature = 'MA'

    def __init__(self,
                 signature_str: str,
                 input_data_array: List[Event],
                 warmup_length: int) -> None:
        self._signal_name = input_data_array[0].name() + "_MA_" + str(warmup_length)
        self._signal_event = Event(self._signal_name)

    def calculate_signal(self) -> float:
        """
        Compute the moving average.
        """
        prices = self.get_data_by_name(self._input_data_array[0].name())
        return sum(prices) / len(prices)


class WMomSignal(SignalBase):
    """
    Weighted momentum signal.
    """

    _signal_signature = 'WM'

    def __init__(self,
                 signature_str: str,
                 input_data_array: List[Event],
                 warmup_length: int) -> None:
        self._signal_name = input_data_array[0].name() + "_WM"
        self._signal_event = Event(self._signal_name)
        # Constants
        self._trading_days_in_month = 21
        self._normalization_factor = 4.0
        self._pivot_months = [1, 3, 6, 12]

    def calculate_signal(self) -> float:
        """
        Compute the weighted momentum.
        """
        prices = self.get_data_by_name(self._input_data_array[0].name())
        weighted_momentum = 0.0
        for _month in self._pivot_months:
            p0 = prices[-1]
            pt = prices[len(prices) - (_month * self._trading_days_in_month) - 1]
            weighted_momentum += (12.0 / _month) * (p0 / pt - 1.0) / self._normalization_factor
        return weighted_momentum


class DataSlot:
    """
    Class for data slot.
    """

    def __init__(self,
                 data_name: str,
                 parent_signals: List[SignalBase]) -> None:
        """
        Initialize data slot.

        Parameters
        ----------
        data_name: str
            Name of the input data.
        parent_signals: List[SignalBase]
            Signals that listens to this data.
        """
        self._data_name = data_name
        self._parent_signals = parent_signals

    def on_event(self, new_data: datacontainer.TimeDouble) -> None:
        """
        Perform the action upon getting an event.

        When there is an event (arrival of data) we want to
        - Update data storage
        - Calculate signal if warming up is complete

        Parameters
        ----------
        new_data: TimeDouble
            Incoming new data.
        """
        for _parent in self._parent_signals:
            # 1. Update data storage
            _parent.update_data(new_data)

            # 2. If warming up is complete and all data arrived, calculate and publish signal
            if len(_parent.get_data_by_name(new_data.get_name())) == \
                    _parent.get_warmup_length() and _parent.check_all_received(new_data.get_name()):
                calculated_signal = _parent.calculate_signal()
                latest_timestamp = _parent.get_time_by_name(new_data.get_name())[-1]
                _parent.get_signal_event().emit(
                    datacontainer.TimeDouble(_parent.get_signal_name(), latest_timestamp, calculated_signal))



#!/usr/bin/env python

"""Tests for `alpyen` package."""

from eventkit import Event
import os
import pytest
import statistics
from typing import List, Dict

from click.testing import CliRunner

from alpyen import datacontainer
from alpyen import backtesting
from alpyen import brokerinterface
from alpyen import cli
from alpyen import livetrading
from alpyen import signal
from alpyen import strategy
from alpyen import utils


@pytest.fixture
def response():
    """Sample pytest fixture.

    See more at: http://doc.pytest.org/en/latest/fixture.html
    """
    # import requests
    # return requests.get('https://github.com/audreyr/cookiecutter-pypackage')


def test_content(response):
    """Sample pytest test function with the pytest fixture as an argument."""
    # from bs4 import BeautifulSoup
    # assert 'GitHub' in BeautifulSoup(response.content).title.string


def test_command_line_interface():
    """Test the CLI."""
    runner = CliRunner()
    result = runner.invoke(cli.main)
    assert result.exit_code == 0
    assert 'alpyen.cli.main' in result.output
    help_result = runner.invoke(cli.main, ['--help'])
    assert help_result.exit_code == 0
    assert '--help  Show this message and exit.' in help_result.output


def test_backtesting_macrossing_reshuffle():
    # Read data
    data_folder = 'Data\\'
    ticker_name = 'BBH'
    file_path = os.path.join(os.path.dirname(__file__), data_folder)
    short_lookback = 5
    long_lookback = 200
    short_lookback_name = ticker_name + '_MA_' + str(short_lookback)
    long_lookback_name = ticker_name + '_MA_' + str(long_lookback)
    ticker_names = [ticker_name]
    all_input = datacontainer.DataUtils.aggregate_yahoo_data(ticker_names, file_path)

    # Subscribe to signals
    signal_info_dict = {}
    signal_info_dict[short_lookback_name]\
        = utils.SignalInfo('MA', ticker_names, [], [], short_lookback, {})
    signal_info_dict[long_lookback_name]\
        = utils.SignalInfo('MA', ticker_names, [], [], long_lookback, {})

    # Subscribe to strategies
    strategy_info_dict = {}
    strategy_name = ticker_name + '_MACrossing_01'
    strategy_info_dict[strategy_name] = utils.StrategyInfo(
        'MACrossing',
        [short_lookback_name, long_lookback_name],
        1, {}, ticker_names, combo_definition={'combo1': [1.0]})

    # Create backtester
    number_path = 4000
    my_backtester = backtesting.Backtester(all_input, ticker_names, signal_info_dict, strategy_info_dict,
                                           number_path)
    my_backtester.run_backtest()
    backtest_results = my_backtester.get_results()

    # Check
    # Actual historical path
    assert backtest_results[strategy_name][str(backtesting.MetricType.PoorMansSharpeRatio)][0]\
           == pytest.approx(0.09503, 0.0001)
    assert backtest_results[strategy_name][str(backtesting.MetricType.MaximumDrawDown)][0]\
           == pytest.approx(0.11913, 0.0001)
    assert backtest_results[strategy_name][str(backtesting.MetricType.Return)][0]\
           == pytest.approx(0.74978, 0.0001)
    # All (including simulated) paths
    assert statistics.mean(backtest_results[strategy_name][str(backtesting.MetricType.PoorMansSharpeRatio)])\
           == pytest.approx(0.105, 0.05)
    assert statistics.stdev(backtest_results[strategy_name][str(backtesting.MetricType.PoorMansSharpeRatio)])\
           == pytest.approx(0.0308, 0.10)
    assert statistics.mean(backtest_results[strategy_name][str(backtesting.MetricType.MaximumDrawDown)])\
           == pytest.approx(0.152, 0.05)
    assert statistics.stdev(backtest_results[strategy_name][str(backtesting.MetricType.MaximumDrawDown)])\
           == pytest.approx(0.0611, 0.05)
    assert statistics.mean(backtest_results[strategy_name][str(backtesting.MetricType.Return)])\
           == pytest.approx(0.865, 0.05)
    assert statistics.stdev(backtest_results[strategy_name][str(backtesting.MetricType.Return)])\
           == pytest.approx(0.326, 0.05)


def test_backtesting_macrossing_resample():
    # Read data
    data_folder = 'Data\\'
    ticker_name = 'BBH'
    file_path = os.path.join(os.path.dirname(__file__), data_folder)
    short_lookback = 5
    long_lookback = 200
    short_lookback_name = ticker_name + '_MA_' + str(short_lookback)
    long_lookback_name = ticker_name + '_MA_' + str(long_lookback)
    ticker_names = [ticker_name]
    all_input = datacontainer.DataUtils.aggregate_yahoo_data(ticker_names, file_path)

    # Subscribe to signals
    signal_info_dict = {}
    signal_info_dict[short_lookback_name]\
        = utils.SignalInfo('MA', ticker_names, [], [], short_lookback, {})
    signal_info_dict[long_lookback_name]\
        = utils.SignalInfo('MA', ticker_names, [], [], long_lookback, {})

    # Subscribe to strategies
    strategy_info_dict = {}
    strategy_name = ticker_name + '_MACrossing_01'
    strategy_info_dict[strategy_name] = utils.StrategyInfo(
        'MACrossing',
        [short_lookback_name, long_lookback_name],
        1, {}, ticker_names, combo_definition={'combo1': [1.0]})

    # Create backtester
    number_path = 4000
    my_backtester = backtesting.Backtester(all_input, ticker_names, signal_info_dict, strategy_info_dict,
                                           number_path)
    my_backtester.run_backtest(backtesting.PathGenerationType.ReturnResampling)
    backtest_results = my_backtester.get_results()

    # Check
    # Actual historical path
    assert backtest_results[strategy_name][str(backtesting.MetricType.PoorMansSharpeRatio)][0]\
        == pytest.approx(0.09503, 0.0001)
    assert backtest_results[strategy_name][str(backtesting.MetricType.MaximumDrawDown)][0]\
        == pytest.approx(0.11913, 0.0001)
    assert backtest_results[strategy_name][str(backtesting.MetricType.Return)][0]\
        == pytest.approx(0.74978, 0.0001)
    # All (including simulated) paths
    assert statistics.mean(backtest_results[strategy_name][str(backtesting.MetricType.PoorMansSharpeRatio)])\
        == pytest.approx(0.105, 0.05)
    assert statistics.stdev(backtest_results[strategy_name][str(backtesting.MetricType.PoorMansSharpeRatio)])\
        == pytest.approx(0.0308, 0.10)
    assert statistics.mean(backtest_results[strategy_name][str(backtesting.MetricType.MaximumDrawDown)])\
        == pytest.approx(0.152, 0.05)
    assert statistics.stdev(backtest_results[strategy_name][str(backtesting.MetricType.MaximumDrawDown)])\
        == pytest.approx(0.06088, 0.10)
    assert statistics.mean(backtest_results[strategy_name][str(backtesting.MetricType.Return)])\
        == pytest.approx(0.865, 0.05)
    assert statistics.stdev(backtest_results[strategy_name][str(backtesting.MetricType.Return)])\
        == pytest.approx(0.326, 0.05)


def test_backtesting_vaa():
    # Read data
    data_folder = 'Data\\'
    ticker_names = ['VOO', 'VWO', 'VEA', 'BND', 'SHY', 'IEF', 'LQD']
    file_path = os.path.join(os.path.dirname(__file__), data_folder)
    all_input = datacontainer.DataUtils.aggregate_yahoo_data(ticker_names, file_path)

    # Subscribe to signals
    signal_info_dict = {}
    lookback = 253
    for ticker in ticker_names:
        signal_info_dict[ticker + '_WM_1'] = utils.SignalInfo('WM', [ticker], [], [], lookback, {})

    # Subscribe to strategies
    strategy_info_dict = {}
    strategy_name = '4ETF_VAA_01'
    strategy_info_dict[strategy_name] = utils.StrategyInfo(
        'VAA', [ticker + '_WM_1' for ticker in ticker_names],
        1, {'risk_on_size': 4, 'num_assets_to_hold': 2,
            'breadth_protection_threshold': 1, 'weighting_scheme': strategy.WeightingScheme.Equal},
        ticker_names, combo_definition={'combo1': [1.0] * len(ticker_names)})

    # Create backtester
    number_path = 1
    my_backtester = backtesting.Backtester(all_input, ticker_names, signal_info_dict, strategy_info_dict,
                                           number_path)
    my_backtester.run_backtest()
    backtest_results = my_backtester.get_results()

    # Check
    # Actual historical path
    assert backtest_results[strategy_name][str(backtesting.MetricType.PoorMansSharpeRatio)][0]\
           == pytest.approx(0.08549, 0.0001)
    assert backtest_results[strategy_name][str(backtesting.MetricType.MaximumDrawDown)][0]\
           == pytest.approx(0.08114, 0.0001)
    assert backtest_results[strategy_name][str(backtesting.MetricType.Return)][0]\
           == pytest.approx(1.01538, 0.0001)

# Signal and Strategy for on-the-fly signal and strategy test
class IncreaseDecrease(signal.SignalBase):
    _signal_signature = 'ID'

    def __init__(self,
                 signature_str: str,
                 input_data_array: List[Event],
                 warmup_length: int) -> None:
        if warmup_length <= 1:
            raise ValueError('IncreaseDecrease: warmup_length must be greater than 1.')
        self._signal_name = input_data_array[0].name() + "_ID_" + str(warmup_length)
        self._signal_event = Event(self._signal_name)

    def calculate_signal(self) -> int:
        prices = self.get_data_by_name(self._input_data_array[0].name())
        return 1 if prices[-1] - prices[0] > 0.0 else -1


class IncreaseDecreasePlus(IncreaseDecrease):
    _signal_signature = 'IDP'

    def __init__(self,
                 signature_str: str,
                 input_data_array: List[Event],
                 warmup_length: int,
                 **kwargs) -> None:
        if warmup_length <= 1:
            raise ValueError('IncreaseDecreasePlus: warmup_length must be greater than 1.')
        self._signal_name = input_data_array[0].name() + "_IDP_" + str(warmup_length)
        self._signal_event = Event(self._signal_name)
        self._threshold = kwargs['threshold']

    def calculate_signal(self) -> int:
        prices = self.get_data_by_name(self._input_data_array[0].name())
        if prices[-1] - prices[0] > self._threshold:
            return 1
        elif prices[-1] - prices[0] < -self._threshold:
            return -1
        else:
            return 0


class BuyIncreaseSellDecrease(strategy.StrategyBase):
    _strategy_signature = 'BISD'

    def __init__(self,
                 signature_str: str,
                 input_signal_array: List[Event],
                 trade_combo: strategy.TradeCombos,
                 warmup_length: int,
                 initial_capital: float = 100.0,
                 order_manager: brokerinterface.OrderManagerBase = None) -> None:
        self._strategy_name = 'BuyIncreaseSellDecrease'

    def make_order_decision(self) -> Dict[str, float]:
        signal_name = next(iter(self._signal_storage))
        if self._signal_storage[signal_name][-1] > 0 and not self.is_currently_long('combo1'):
            return {'combo1': 0.2, 'combo2': -0.2}
        elif self._signal_storage[signal_name][-1] < 0 and not self.is_currently_long('combo2'):
            return {'combo1': -0.2, 'combo2': 0.2}


class BuyIncreaseSellDecreasePlus(BuyIncreaseSellDecrease):
    _strategy_signature = 'BISDP'

    def __init__(self,
                 signature_str: str,
                 input_signal_array: List[Event],
                 trade_combo: strategy.TradeCombos,
                 warmup_length: int,
                 initial_capital: float = 100.0,
                 order_manager: brokerinterface.OrderManagerBase = None) -> None:
        self._strategy_name = 'BuyIncreaseSellDecreasePlus'

    def make_order_decision(self) -> Dict[str, float]:
        signal_name = next(iter(self._signal_storage))
        weight_1 = 0.0
        weight_2 = 0.0
        if self._signal_storage[signal_name][-1] > 0:
            if self.is_currently_long('combo1'):
                weight_1 = 0.01
            else:
                weight_1 = 0.2
        else:
            weight_1 = -0.22
        if self._signal_storage[signal_name][-1] < 0:
            if self.is_currently_long('combo2'):
                weight_2 = 0.01
            else:
                weight_2 = 0.2
        else:
            weight_2 = -0.22
        return {'combo1': weight_1, 'combo2': weight_2}


def test_backtesting_on_the_fly_signal_strategy():
    # Read data
    data_folder = 'Data\\'
    signal_ticker_name = 'VWO'
    trade_ticker_names = ['VOO', 'SHY']
    file_path = os.path.join(os.path.dirname(__file__), data_folder)
    warmup_length = 5
    signal_name1 = signal_ticker_name + '_ID_' + str(warmup_length)
    signal_name2 = signal_ticker_name + '_IDP_' + str(warmup_length)
    signal_ticker_names = [signal_ticker_name]
    all_input = datacontainer.DataUtils.aggregate_yahoo_data(signal_ticker_names + trade_ticker_names, file_path)

    # Subscribe to signals
    signal_info_dict = {}
    signal_info_dict[signal_name1]\
        = utils.SignalInfo('ID', signal_ticker_names, [], [], warmup_length, {})
    signal_info_dict[signal_name2]\
        = utils.SignalInfo('IDP', signal_ticker_names, [], [], warmup_length, {'threshold': 0.05})

    # Subscribe to strategies
    strategy_info_dict = {}
    strategy_name = signal_ticker_name + '_BuyIncreaseSellDecrease'
    strategy_info_dict[strategy_name] = utils.StrategyInfo(
        'BISD',
        [signal_name1],
        1, {}, trade_ticker_names, combo_definition={'combo1': [1.0, -3.0], 'combo2': [-1.0, 2.0]})
    strategy_info_dict[strategy_name+'Plus'] = utils.StrategyInfo(
        'BISDP',
        [signal_name2],
        1, {}, trade_ticker_names, combo_definition={'combo1': [1.0, -3.0], 'combo2': [-1.0, 2.0]})

    # Create backtester
    number_path = 1
    my_backtester = backtesting.Backtester(all_input, trade_ticker_names + signal_ticker_names,
                                           signal_info_dict, strategy_info_dict,
                                           number_path)
    my_backtester.run_backtest()
    backtest_results = my_backtester.get_results()

    # Check
    # Actual historical path
    assert backtest_results[strategy_name][str(backtesting.MetricType.PoorMansSharpeRatio)][0]\
        == pytest.approx(0.015934, 0.0001)
    assert backtest_results[strategy_name][str(backtesting.MetricType.MaximumDrawDown)][0]\
        == pytest.approx(0.39325, 0.0001)
    assert backtest_results[strategy_name][str(backtesting.MetricType.Return)][0]\
        == pytest.approx(0.266176, 0.0001)
    assert backtest_results[strategy_name+'Plus'][str(backtesting.MetricType.PoorMansSharpeRatio)][0]\
        == pytest.approx(-0.013084, 0.0001)
    assert backtest_results[strategy_name+'Plus'][str(backtesting.MetricType.MaximumDrawDown)][0]\
        == pytest.approx(1.05596, 0.0001)
    assert backtest_results[strategy_name+'Plus'][str(backtesting.MetricType.Return)][0]\
        == pytest.approx(6.1448, 0.0001)


# Signal and Strategy for debugging live trading
class TempSignal(signal.SignalBase):
    _signal_signature = 'TS'

    def __init__(self,
                 signature_str: str,
                 input_data_array: List[Event],
                 warmup_length: int,
                 **kwargs) -> None:
        self._signal_name = input_data_array[0].name() + "_TS_" + str(warmup_length)
        self._signal_event = Event(self._signal_name)
        self._num_sig_fig = kwargs['num_sig_fig']

    def calculate_signal(self) -> int:
        prices = self.get_data_by_name(self._input_data_array[0].name())
        return round(prices[-1] / prices[-2] * (10.0 ** self._num_sig_fig), 0) % 10


class TempStrategy(strategy.StrategyBase):
    _strategy_signature = 'TempStrategy'

    def __init__(self,
                 signature_str: str,
                 input_signal_array: List[Event],
                 trade_combo: strategy.TradeCombos,
                 warmup_length: int,
                 initial_capital: float = 100.0,
                 order_manager: brokerinterface.OrderManagerBase = None) -> None:
        self._strategy_name = 'My temp strategy'

    def make_order_decision(self) -> Dict[str, float]:
        signal_name = next(iter(self._signal_storage))
        if 2 > self._signal_storage[signal_name][-1] >= 0:
            return {'combo_1': 2.0}
        elif 4 > self._signal_storage[signal_name][-1] >= 2:
            return {'combo_1': -2.0}
        elif 6 > self._signal_storage[signal_name][-1] >= 4:
            return {'combo_2': 3.0}
        elif 8 > self._signal_storage[signal_name][-1] >= 6:
            return {'combo_2': -3.0}
        else:
            return {'combo_3': 1.0}


def test_backtesting_traded_contracts_not_in_signal():
    # In this scenario, some of the traded contracts are not used in any signal

    # Read data
    data_folder = 'Data\\'
    input_tickers = ['BBH']
    trade_tickers = ['VWO', 'VEA', 'BND']
    file_path = os.path.join(os.path.dirname(__file__), data_folder)
    all_input = datacontainer.DataUtils.aggregate_yahoo_data(input_tickers + trade_tickers, file_path)

    # Subscribe to signals
    signal_info_dict = {}
    for ticker in input_tickers:
        signal_info_dict[ticker + '_TS_1'] = utils.SignalInfo('TS', [ticker],
                                                              [utils.ContractType.FX],
                                                              [utils.PriceOHLCType.Close], 2, {'num_sig_fig': 5})

    # Subscribe to strategies
    strategy_info_dict = {}
    strategy_name = 'Temp_Strategy'
    trade_combo = {'combo_1': [-0.3, 0.6, 0.0],
                   'combo_2': [-0.6, 0.0, -0.2],
                   'combo_3': [0.0, -0.6, 0.2]}
    strategy_info_dict[strategy_name] = utils.StrategyInfo(
        'TempStrategy',
        [ticker + '_TS_1' for ticker in input_tickers],
        1,
        {},
        trade_tickers,
        [utils.ContractType.FX] * 3,
        trade_combo)

    # Create backtester
    number_path = 1
    my_backtester = backtesting.Backtester(all_input, input_tickers, signal_info_dict,
                                           strategy_info_dict, number_path)
    my_backtester.run_backtest()
    backtest_results = my_backtester.get_results()

    # Check
    # Actual historical path
    assert backtest_results[strategy_name][str(backtesting.MetricType.PoorMansSharpeRatio)][0]\
           == pytest.approx(0.04791, 0.0001)
    assert backtest_results[strategy_name][str(backtesting.MetricType.MaximumDrawDown)][0]\
           == pytest.approx(0.69536, 0.0001)
    assert backtest_results[strategy_name][str(backtesting.MetricType.Return)][0]\
           == pytest.approx(0.42129, 0.0001)


def test_live_trading():
    # Subscribe to signals
    input_tickers = ['USDJPY']
    trade_tickers = ['EURUSD', 'GBPUSD']
    signal_info_dict = {}

    warmup_length = 2
    signal_name = input_tickers[0] + '_ID_' + str(warmup_length)
    signal_info_dict[signal_name]\
        = utils.SignalInfo('ID',
                           input_tickers,
                           [utils.ContractType.FX],
                           [utils.PriceOHLCType.Close],
                           warmup_length,
                           {})

    # Subscribe to strategies
    strategy_info_dict = {}
    strategy_name = 'BISD'
    trade_combo = {'combo1': [1000.0, -3000.0], 'combo2': [-1000.0, 2000.0]}
    strategy_info_dict[strategy_name] = utils.StrategyInfo(
        'BISD',
        [signal_name],
        1,
        {},
        trade_tickers,
        [utils.ContractType.FX] * 2,
        trade_combo)

    # Create live trader
    path_to_control_file = 'test_control.yml'
    my_trader = livetrading.LiveTrader('IB', signal_info_dict, strategy_info_dict, path_to_control_file)
    my_trader.start_trading()


def test_live_trading_2():
    # Subscribe to signals
    input_tickers = ['USDJPY']
    trade_tickers = ['EURUSD', 'GBPUSD', 'CHFUSD']
    signal_info_dict = {}

    for ticker in input_tickers:
        signal_info_dict[ticker + '_TS_1'] = utils.SignalInfo('TS', [ticker],
                                                              [utils.ContractType.FX],
                                                              [utils.PriceOHLCType.Close], 2, {'num_sig_fig': 5})

    # Subscribe to strategies
    strategy_info_dict = {}
    strategy_name = 'Temp_Strategy'
    trade_combo = {'combo_1': [-50.0, 100.0, 0.0],
                   'combo_2': [-100.0, 0.0, -60.0],
                   'combo_3': [0.0, -100.0, 70.0]}
    strategy_info_dict[strategy_name] = utils.StrategyInfo(
        'TempStrategy',
        [ticker + '_TS_1' for ticker in input_tickers],
        1,
        {},
        trade_tickers,
        [utils.ContractType.FX] * 3,
        trade_combo)

    # Create live trader
    path_to_control_file = 'test_control.yml'
    my_trader = livetrading.LiveTrader('IB', signal_info_dict, strategy_info_dict, path_to_control_file)
    my_trader.start_trading()



import enum
from eventkit import Event
from itertools import accumulate
import random
from typing import List, Dict

from . import datacontainer
from . import signal
from . import strategy
from . import utils


class PathGenerationType(enum.Enum):
    """
    Enum class for path generation type.
    """
    ReturnShuffling = 1
    ReturnResampling = 2


class MetricType(enum.Enum):
    """
    Enum class for performance metric.
    """
    PoorMansSharpeRatio = 1  # Assuming zero risk free rate
    MaximumDrawDown = 2
    Return = 3


class Backtester:
    """
    Class for backtester.
    """
    def __init__(self,
                 aggregated_input: List[datacontainer.PriceTimeSeries],
                 names: List[str],
                 signal_info: Dict[str, utils.SignalInfo],
                 strategy_info: Dict[str, utils.StrategyInfo],
                 number_simulation: int = 1) -> None:
        """
        Initialize backtester.

        Parameters
        ----------
        aggregated_input: List[PriceTimeSeries]
            All input data used for backtesting.
        names: List[str]
            Data event names.
        signal_info: Dict[str, utils.SignalInfo]
            Information for building signals.
        strategy_info: Dict[str, utils.StrategyInfo]
            Information for building strategies.
        number_simulation: int
            Number of simulation paths.
        """
        self._aggregated_input = aggregated_input
        self._names = names
        self._signal_info = signal_info
        self._strategy_info = strategy_info
        self._number_simulation = number_simulation
        self._results: Dict[str, Dict[str, List[float]]] = {}

    def run_backtest(self,
                     path_type: PathGenerationType = PathGenerationType.ReturnShuffling) -> None:
        """
        Run backtest.

        Parameters
        ----------
        path_type: PathGenerationType
            Path generation type.
        """
        # Build dictionaries to store simulation results
        simulation_results_dict: Dict[str, Dict[str, List[float]]] = {}
        for key in self._strategy_info.keys():
            simulation_results_dict[key] = {}

        for sim_i in range(self._number_simulation):
            # Create incoming data
            data_event_dict: Dict[str, Event] = {}
            for name in self._names:
                data_event_dict[name] = Event(name)
            # In case there are traded contracts that are not in any signal
            for k, v in self._strategy_info.items():
                for i in range(len(v.get_contract_names())):
                    if v.get_contract_names()[i] not in data_event_dict:
                        data_event_dict[v.get_contract_names()[i]] = Event(v.get_contract_names()[i])

            # Create signals
            signal_dict = self.create_signal_dict(data_event_dict)

            # Create strategies
            strategy_dict = self.create_strategy_dict(signal_dict, data_event_dict)

            # Read and fire data; the first trial is always the actual history
            price_dict: Dict[str, List[float]] = {}
            for j in range(len(self._aggregated_input)):
                if sim_i == 0:
                    price_dict[self._aggregated_input[j].get_name()] = self._aggregated_input[j].get_price()
                else:
                    price_dict[self._aggregated_input[j].get_name()] = \
                        self.generate_simulated_path(self._aggregated_input[j].get_price(), path_type)

            for t in range(len(price_dict[next(iter(price_dict))])):
                for k, v in data_event_dict.items():
                    data_name = k
                    timestamp = self._aggregated_input[0].get_timestamp()[t]
                    price = price_dict[data_name][t]
                    v.emit(datacontainer.TimeDouble(data_name, timestamp, price))

            # Save results
            for k, v in strategy_dict.items():
                simulation_results_dict[k][k + str(sim_i)] = v.get_mtm_history()
        # Analyze results and print out
        metrics_to_calculate = [MetricType.PoorMansSharpeRatio,
                                MetricType.MaximumDrawDown,
                                MetricType.Return]
        for key in self._strategy_info.keys():
            metrics = self.calculate_metrics(simulation_results_dict[key], metrics_to_calculate)
            self._results[key] = metrics

    def get_results(self):
        return self._results

    def create_signal_dict(self,
                           data_event_dict: Dict[str, Event]) -> Dict[str, signal.SignalBase]:
        """
        Create signal dictionary.

        Parameters
        ----------
        data_event_dict: Dict[str, Event]
            Data event dictionary.
        """
        output_dict: Dict[str, signal.SignalBase] = {}
        signal_info_dict: Dict[str, utils.SignalInfo] = self._signal_info
        for k, v in signal_info_dict.items():
            price_event_list: List[Event] = []
            for name_i in v.get_input_names():
                price_event_list.append(data_event_dict.get(name_i))
            # Create signal
            output_dict[k] = signal.SignalBase(v.get_signal_signature(),
                                               price_event_list,
                                               v.get_warmup_length(),
                                               **v.get_custom_params())
        return output_dict

    def create_strategy_dict(self,
                             signal_dict: Dict[str, signal.SignalBase],
                             data_event_dict: Dict[str, Event]) -> Dict[str, strategy.StrategyBase]:
        """
        Create strategy dictionary.

        Parameters
        ----------
        signal_dict: Dict[str, signal.SignalBase]
            Signal dictionary.
        data_event_dict: Dict[str, Event]
            Data event dictionary.
        """
        output_dict: Dict[str, strategy.StrategyBase] = {}
        strategy_info_dict: Dict[str, utils.StrategyInfo] = self._strategy_info
        for k, v in strategy_info_dict.items():
            price_event_list: List[Event] = []
            for i in range(len(v.get_contract_names())):
                price_event_list.append(data_event_dict.get(v.get_contract_names()[i]))
            # Create TradeCombos
            trade_combos = strategy.TradeCombos(price_event_list, v.get_combo_definition())
            # Create strategy
            signal_event_list: List[Event] = []
            for i in range(len(v.get_input_names())):
                signal_event_list.append(signal_dict.get(v.get_input_names()[i]).get_signal_event())
            output_dict[k] = strategy.StrategyBase(v.get_strategy_signature(), signal_event_list,
                                                   trade_combos, v.get_warmup_length(), **v.get_custom_params())
        return output_dict

    def generate_simulated_path(self,
                                original_data: List[float],
                                path_type: PathGenerationType = PathGenerationType.ReturnShuffling) -> List[float]:
        """
        Generate simulated path from original price data.

        Parameters
        ----------
        original_data: List[float]
            Original price data array.
        path_type: PathGenerationType
            Path generation type.

        Returns
        -------
            List[float]
                Simulated price data.
        """
        if path_type == PathGenerationType.ReturnShuffling:
            return self.generate_shuffled_return_path(original_data)
        elif path_type == PathGenerationType.ReturnResampling:
            return self.generate_resampled_return_path(original_data)
        else:
            raise ValueError('Backtester.generate_simulated_path: Unknown generation type.')

    def generate_shuffled_return_path(self,
                                      original_data: List[float]) -> List[float]:
        """
        Generate shuffled path from original price data.

        Parameters
        ----------
        original_data: List[float]
            Original price data array.

        Returns
        -------
            List[float]
                Simulated price data.
        """
        # Calculate return from price
        return_list = [p_t / p_tm1 - 1 for p_t, p_tm1 in zip(original_data[1:], original_data[:-1])]

        # Shuffle
        random.shuffle(return_list)

        # Create new price series
        price_t0 = original_data[0]
        return_list = [price_t0] + return_list
        new_data = [*accumulate(return_list, lambda a, b: a * (1.0 + b))]
        return new_data

    def generate_resampled_return_path(self,
                                       original_data: List[float]) -> List[float]:
        """
        Generate resampled path from original price data.

        Parameters
        ----------
        original_data: List[float]
            Original price data array.

        Returns
        -------
            List[float]
                Simulated price data.
        """
        # Calculate return from price
        return_list = [p_t / p_tm1 - 1 for p_t, p_tm1 in zip(original_data[1:], original_data[:-1])]

        # Sample
        random_indices = list(range(0, len(return_list), 1))
        random.shuffle(random_indices)
        return_randomized = [return_list[i] for i in random_indices]

        # Create new price series
        price_t0 = original_data[0]
        return_randomized = [price_t0] + return_randomized
        new_data = [*accumulate(return_randomized, lambda a, b: a * (1.0 + b))]
        return new_data

    def calculate_metrics(self,
                          simulation_results: Dict[str, List[float]],
                          metrics_to_calculate: List[MetricType]) -> Dict[str, List[float]]:
        """
        Calculate a set of specifiec metrics.

        Parameters
        ----------
        simulation_results: Dict[str, List[float]]
            Simulation results.
        metrics_to_calculate: List[MetricType]
            Entities to calculate

        Returns
        -------
            Dict[str, List[float]]
                Metrics.
        """
        output = {}

        for i in range(len(metrics_to_calculate)):
            metric_storage = []
            for k, v in simulation_results.items():
                return_list = [p_t / p_tm1 - 1 for p_t, p_tm1 in zip(v[1:], v[:-1])]
                price_list = v
                if metrics_to_calculate[i] == MetricType.PoorMansSharpeRatio:
                    metric_storage.append(self.calculate_poor_mans_sharpe_ratio(return_list))
                elif metrics_to_calculate[i] == MetricType.MaximumDrawDown:
                    metric_storage.append(self.calculate_mdd(price_list))
                elif metrics_to_calculate[i] == MetricType.Return:
                    metric_storage.append(price_list[-1] / price_list[0] - 1.0)
                else:
                    raise ValueError('Backtester.calculate_metrics: Unknown metrics type.')
            output[str(metrics_to_calculate[i])] = metric_storage
        return output

    def calculate_poor_mans_sharpe_ratio(self,
                                         return_list: List[float]) -> float:
        """
        Calculate 'poor man's Sharpe ratio' (i.e. assuming risk-free rate is zero)

        Parameters
        ----------
        return_list: List[float]
            Return series.

        Returns
        -------
            float
                Poor man's Sharpe ratio.
        """
        if len(return_list) == 0:
            return 0.0
        else:
            average = sum(return_list) / len(return_list)
            deviation_squared = [(x - average)**2 for x in return_list]
            return average / (sum(deviation_squared) / (len(deviation_squared) - 1))**0.5

    def calculate_mdd(self,
                      price_list: List[float]) -> float:
        """
        Calculate maximum drawdown

        Parameters
        ----------
        price_list: List[float]
            Price series.

        Returns
        -------
            float
                Maximum drawdown.
        """
        mdd = 0.0
        peak = price_list[0]
        for p_i in price_list:
            if p_i > peak:
                peak = p_i
            dd = (peak - p_i) / peak
            if dd > mdd:
                mdd = dd
        return mdd
